"""GUI for selecting topographical components"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>

# Document:  represents data
# ChangeAction:  modifies Document
# Model:  creates ChangeActions and applies them to the History
# Frame:
#  - visualizaes Document
#  - listens to Document changes
#  - issues commands to Model
from collections import defaultdict
from itertools import repeat
from math import ceil
from operator import itemgetter
import re
from typing import Sequence, Tuple, Union

import mne
import matplotlib.figure
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_agg import FigureCanvasAgg
import numpy as np
from scipy import linalg
import seaborn
import wx
from wx.lib.scrolledpanel import ScrolledPanel

from .. import load, plot, fmtxt
from .._colorspaces import UNAMBIGUOUS_COLORS
from .._data_obj import Dataset, Factor, NDVar, Categorial, Scalar, combine
from .._io.fiff import _picks
from .._ndvar import concatenate, neighbor_correlation
from .._types import PathArg
from .._utils.numpy_utils import INT_TYPES
from .._utils.parse import FLOAT_PATTERN, POS_FLOAT_PATTERN
from .._utils.system import IS_OSX
from ..mne_fixes._version import MNE_VERSION, V0_24
from ..plot._base import DISPLAY_UNIT, UNIT_FORMAT, AxisData, DataLayer, PlotType
from ..plot._topo import AxTopomap
from .frame import EelbrainDialog
from .history import Action, FileDocument, FileModel, FileFrame, FileFrameChild
from .mpl_canvas import FigureCanvasPanel
from .text import HTML2Frame as HTMLFrame
from .utils import Icon, REValidator
from . import ID


COLOR = {True: (.5, 1, .5), False: (1, .3, .3)}
LINE_COLOR = {True: 'k', False: (1, 0, 0)}
TOPO_ARGS = {
    'interpolation': 'linear',  # interpolation that does not assume continuity
    'clip': 'even',
}

# For unit-tests
TEST_MODE = False


class ChangeAction(Action):
    """Action objects are kept in the history and can do and undo themselves

    Parameters
    ----------
    desc : str
        Description of the action
        list of (i, name, old, new) tuples
    """
    def __init__(self, desc, index=None, old_accept=None, new_accept=None,
                 old_path=None, new_path=None):
        self.desc = desc
        self.index = index
        self.old_path = old_path
        self.old_accept = old_accept
        self.new_path = new_path
        self.new_accept = new_accept

    def do(self, doc):
        if self.index is not None:
            doc.set_case(self.index, self.new_accept)
        if self.new_path is not None:
            doc.set_path(self.new_path)

    def undo(self, doc):
        if self.index is not None:
            doc.set_case(self.index, self.old_accept)
        if self.new_path is not None and self.old_path is not None:
            doc.set_path(self.old_path)


class Document(FileDocument):
    """Represents data for the current state of the Document

    (Data can be accessed, but should only be modified through the Model)

    Parameters
    ----------
    path
        Path to the ICA file.
    data
        Dataset containing 'epochs' (mne Epochs), 'index' (Var describing
        epochs) and variables describing cases in epochs, used to plot
        condition averages.
    """
    def __init__(
            self,
            path: PathArg,
            data: Union[Dataset, mne.BaseEpochs],
            sysname: str,
            adjacency: Union[str, Sequence] = None,
            drop_epochs_std: float = None,  # drop epochs with high signal (e.g. 10)
            decim: int = None,
    ):
        FileDocument.__init__(self, path)
        self._ndvar_args = dict(sysname=sysname, adjacency=adjacency)
        self.saved = True
        self._explained_variance = {}

        if isinstance(data, mne.io.BaseRaw):
            events = mne.make_fixed_length_events(data)
            if decim is None:
                decim = int(round(data.info['sfreq'] / 100))
            ds = Dataset({'epochs': mne.Epochs(data, events, 1, 0, 1, baseline=None, proj=False, decim=decim, preload=True)})
        elif isinstance(data, mne.BaseEpochs):
            ds = Dataset({'epochs': data})
        elif isinstance(data, Dataset):
            ds = data
        else:
            raise TypeError(f'{data=}')

        # Exclude extreme epochs
        epochs_ndvar = self.as_ndvar(ds['epochs'])
        if drop_epochs_std:
            variance = epochs_ndvar.var('sensor').max('time')
            index = np.arange(len(variance))
            while True:
                sub_variance = variance[index]
                sub_variance -= sub_variance.mean()
                sub_variance /= sub_variance.std()
                print(f"Max variance: {sub_variance.max()} STD...")
                if sub_variance.max() > drop_epochs_std:
                    mask = sub_variance <= drop_epochs_std
                    print(f"Excluding {np.sum(~mask)} epochs > {drop_epochs_std} STD: {index[~mask]}...")
                    index = index[mask]
                else:
                    break
            ds = ds[index]
            epochs_ndvar = epochs_ndvar[index]

        # Read ICA
        self.ica = ica = mne.preprocessing.read_ica(path)
        self.accept = np.ones(self.ica.n_components_, bool)
        self.accept[ica.exclude] = False
        self.epochs = epochs = ds['epochs']
        self.epochs_ndvar = epochs_ndvar
        self.ds = ds
        # for 3d-data, pick magnetometers
        picks = _picks(ica.info, None, 'bads')

        # components
        data = np.dot(ica.mixing_matrix_.T, ica.pca_components_[:ica.n_components_])
        ic_dim = Scalar('component', np.arange(len(data)))
        self.components = NDVar(data[:, picks], (ic_dim, self.epochs_ndvar.sensor), 'components', {'meas': 'component', 'cmap': 'xpolar'})

        # sources
        data = ica.get_sources(epochs).get_data(copy=False)
        self.sources = NDVar(data, ('case', ic_dim, self.epochs_ndvar.time), 'sources', {'meas': 'component', 'cmap': 'xpolar'})

        # find unique epoch labels
        if 'index' in ds:
            labels = map(str, ds['index'])
            if 'epoch' in ds:
                labels = map(' '.join, zip(ds['epoch'], labels))
        else:
            labels = map(str, range(len(epochs)))
        self.epoch_labels = tuple(labels)

        # properties which are not modified by ICA
        # global mean
        if ica.noise_cov is None:  # revert standardization
            global_mean = ica.pca_mean_ * ica.pre_whitener_[:, 0]
        else:
            global_mean = np.dot(linalg.pinv(ica.pre_whitener_), ica.pca_mean_)
        self.global_mean = NDVar(global_mean[picks], (self.epochs_ndvar.sensor,))
        # pre-ICA signal range
        self.pre_ica_min = self.epochs_ndvar.min('sensor')
        self.pre_ica_max = self.epochs_ndvar.max('sensor')
        self.pre_ica_range_scale = (self.pre_ica_max.mean() - self.pre_ica_min.mean())
        self.pre_ica_min /= self.pre_ica_range_scale
        self.pre_ica_max /= self.pre_ica_range_scale

        # publisher
        self.callbacks.register_key('case_change')

    def as_ndvar(self, epochs):
        if isinstance(epochs, mne.BaseEpochs):
            return load.mne.epochs_ndvar(epochs, **self._ndvar_args)
        elif isinstance(epochs, mne.Evoked):
            return load.mne.evoked_ndvar(epochs, **self._ndvar_args)
        raise TypeError(f"{epochs=}")

    def apply(self, inst):
        if isinstance(inst, list):
            return [self.ica.apply(i.copy()) for i in inst]
        else:
            return self.ica.apply(inst.copy())

    def set_case(self, index, state):
        self.accept[index] = state
        self.ica.exclude = list(np.flatnonzero(np.invert(self.accept)))
        self.callbacks.callback('case_change', index)

    def save(self):
        if MNE_VERSION >= V0_24:
            self.ica.save(self.path, overwrite=True)
        else:
            self.ica.save(self.path)

    def explained_variance(self, component: int, format: bool = False):
        if component not in self._explained_variance:
            self._explained_variance[component] = self.ica.get_explained_variance_ratio(self.epochs, components=component)
        desc_dict = self._explained_variance[component]
        if format:
            if len(desc_dict) == 1:
                return ''.join(f'{v:.1%}' for v in desc_dict.values())
            else:
                return ', '.join([f'{k}: {v:.1%}' for k, v in desc_dict.items()])
        return desc_dict


class Model(FileModel):
    """Manages a document with its history"""

    def __init__(self, doc: Document):
        FileModel.__init__(self, doc)

    def set_case(self, index, state, desc="Manual Change"):
        old_accept = self.doc.accept[index]
        action = ChangeAction(desc, index, old_accept, state)
        self.history.do(action)

    def toggle(self, case):
        old_accept = self.doc.accept[case]
        action = ChangeAction("Manual toggle", case, old_accept, not old_accept)
        self.history.do(action)

    def clear(self):
        action = ChangeAction("Clear", slice(None), self.doc.accept.copy(), True)
        self.history.do(action)


class ContextMenu(wx.Menu):
    "Helper class for Menu to store component ID"
    def __init__(self, i_comp: int = None, i_epoch: int = None):
        wx.Menu.__init__(self)
        self.i_comp = i_comp
        self.i_epoch = i_epoch


class SharedToolsMenu:  # Frame mixin
    # set by FileFrame:
    doc = None
    config = None
    # MakeToolsMenu() might be called before __init__
    butterfly_baseline = ID.BASELINE_NONE
    last_model = ""

    def AddToolbarButtons(self, tb):
        button = wx.Button(tb, label="Rare Events")
        button.Bind(wx.EVT_BUTTON, self.OnFindRareEvents)
        tb.AddControl(button)
        button = wx.Button(tb, label="Noisy Epochs")
        button.Bind(wx.EVT_BUTTON, self.OnFindNoisyEpochs)
        tb.AddControl(button)
        button = wx.Button(tb, label="PSD")
        button.Bind(wx.EVT_BUTTON, self.OnPlotPSD)
        tb.AddControl(button)

    def MakeToolsMenu(self, menu):
        app = wx.GetApp()

        # Artifact detection helpers
        item = menu.Append(wx.ID_ANY, "Find Rare Events", "Find components with major loading on a small number of epochs")
        app.Bind(wx.EVT_MENU, self.OnFindRareEvents, item)
        item = menu.Append(wx.ID_ANY, "Find Noisy Epochs", "Find epochs with strong signal")
        app.Bind(wx.EVT_MENU, self.OnFindNoisyEpochs, item)
        item = menu.Append(wx.ID_ANY, "Find Bad Channels", "Find components that are likely due to bad channels")
        app.Bind(wx.EVT_MENU, self.OnFindBadChannels, item)
        menu.AppendSeparator()

        # plotting
        item = menu.Append(wx.ID_ANY, "Butterfly Plot Grand Average", "Plot the grand average of all epochs")
        app.Bind(wx.EVT_MENU, self.OnPlotGrandAverage, item)
        item = menu.Append(wx.ID_ANY, "Butterfly Plot by Category", "Separate butterfly plots for different model cells")
        app.Bind(wx.EVT_MENU, self.OnPlotButterfly, item)
        # Baseline submenu
        blmenu = wx.Menu()
        blmenu.AppendRadioItem(ID.BASELINE_CUSTOM, "Baseline Period")
        blmenu.AppendRadioItem(ID.BASELINE_GLOABL_MEAN, "Global Mean")
        blmenu.AppendRadioItem(ID.BASELINE_NONE, "No Baseline Correction")
        blmenu.Check(self.butterfly_baseline, True)
        blmenu.Bind(wx.EVT_MENU, self.OnSetButterflyBaseline, id=ID.BASELINE_CUSTOM)
        blmenu.Bind(wx.EVT_MENU, self.OnSetButterflyBaseline, id=ID.BASELINE_GLOABL_MEAN)
        blmenu.Bind(wx.EVT_MENU, self.OnSetButterflyBaseline, id=ID.BASELINE_NONE)
        menu.AppendSubMenu(blmenu, "Baseline")

    def OnFindNoisyEpochs(self, event):
        unit = self.doc.epochs_ndvar.info.get('unit', '<unknown unit>')
        if unit in DISPLAY_UNIT:
            display_unit = DISPLAY_UNIT[unit]
            scale_factor = 1 / UNIT_FORMAT[display_unit]
        else:
            display_unit = unit
            scale_factor = None
        dlg = FindNoisyEpochsDialog(self, unit=display_unit)
        rcode = dlg.ShowModal()
        dlg.Destroy()
        if rcode != wx.ID_OK:
            return
        threshold = float(dlg.threshold.GetValue())
        threshold_desc = f'{threshold:g} {display_unit}'
        if scale_factor:
            threshold *= scale_factor
        apply_rejection = dlg.apply_rejection.GetValue()
        sort_by_component = dlg.sort_by_component.GetValue()
        max_ch_ratio = dlg.max_ch_ratio.GetValue()
        if max_ch_ratio:
            max_ch_ratio = float(max_ch_ratio)
        else:
            max_ch_ratio = 0
        dlg.StoreConfig()

        # compute and rank
        if apply_rejection:
            epochs = self.doc.as_ndvar(self.doc.apply(self.doc.epochs))
        else:
            epochs = self.doc.epochs_ndvar
        peaks = epochs.extrema(('time', 'sensor')).abs().x

        # collect output
        res = [(i, peak) for i, peak in enumerate(peaks) if peak >= threshold]  # epoch, value
        if len(res) == 0:
            wx.MessageBox(f"No epochs with signals exceeding {threshold_desc} were found.", "No Noisy Epochs Found", style=wx.ICON_INFORMATION)
            return

        if sort_by_component:
            res_by_component = defaultdict(list)
            # Find contribution of each component
            component_magnitude = self.doc.components.abs().sum('sensor')
            if apply_rejection:
                component_magnitude.x *= self.doc.accept
            magnitude = self.doc.sources.abs().sum('time') * component_magnitude
            for i, peak in res:
                magnitude_i = magnitude[i]
                c_max = magnitude_i.argmax()
                ratio = magnitude_i[c_max] / magnitude_i.sum()
                res_by_component[c_max].append((i, peak, ratio))
            # Sort epochs by ratio
            for res_list in res_by_component.values():
                res_list.sort(key=itemgetter(2), reverse=True)
            # Sort components by max ratio
            max_ratio = {component: values[0][2] for component, values in res_by_component.items()}
            sorted_components = sorted(max_ratio, key=lambda c: max_ratio[c], reverse=True)
            res_by_component = {c: res_by_component[c] for c in sorted_components}
        else:
            res_by_component = None

        # format output
        doc = fmtxt.Section("Noisy epochs")
        doc.add_paragraph(f"Epochs with signal exceeding {threshold_desc}")
        if sort_by_component:
            doc.add_paragraph("Sorted by dominant component")
        doc.append(fmtxt.linebreak)
        if sort_by_component:
            for component, values in res_by_component.items():
                # test whether this is a single noisy channel
                channel_values = np.sort(np.abs(self.doc.components[component].x))
                max_channel_ratio = channel_values[-1] / channel_values[-2]
                if max_ch_ratio and max_channel_ratio > max_ch_ratio:
                    continue
                # plot component map
                figure = matplotlib.figure.Figure(figsize=(1, 1))
                canvas = FigureCanvasAgg(figure)
                axes = figure.add_subplot()
                plot.Topomap(self.doc.components[component], axes=axes)
                image = fmtxt.Image(f'#{component}', 'jpg')
                canvas.print_jpeg(image)
                # Component properties
                # sec = doc.add_section(f"#{component}")
                sec = doc
                heading = fmtxt.FMTextElement(f"#{component}", 'h2')
                table = fmtxt.Table('lll', rules=False)
                table.cells(image, heading, f'{max_channel_ratio:.1f}')
                sec.add_paragraph(table)
                # sec.add_paragraph([image, f"Ch 1/2 ratio: {max_channel_ratio:.1f}", fmtxt.linebreak])
                # add links to epochs
                by_ratio = defaultdict(list)
                for i, peak, ratio in values:
                    by_ratio[f'{ratio:.0%}'].append(i)
                for ratio, epochs in by_ratio.items():
                    sec.append(f'{ratio}: ')
                    for i in epochs:
                        sec.append([fmtxt.Link(self.doc.epoch_labels[i], f'component:{component} epoch:{i}'), ', '])
                    sec.append(fmtxt.linebreak)
        else:
            for i, peak in res:
                doc.append(fmtxt.Link(self.doc.epoch_labels[i], f'epoch:{i}'))
                doc.append(f": {peak:g}")
                doc.append(fmtxt.linebreak)
        InfoFrame(self, "Noisy Epochs", doc, 300)

    def OnFindRareEvents(self, event):
        dlg = FindRareEventsDialog(self)
        rcode = dlg.ShowModal()
        dlg.Destroy()
        if rcode != wx.ID_OK:
            return
        threshold = float(dlg.threshold.GetValue())
        dlg.StoreConfig()

        # compute and rank
        y = self.doc.sources.max('time') - self.doc.sources.min('time')
        z = (y - y.mean('case')) / y.std('case')
        z_max = z.max('case').x
        components_ranked = np.argsort(z_max)[::-1]

        # collect output
        res = []
        for c in components_ranked:
            if z_max[c] < threshold:
                break
            z_epochs = z.x[:, c]
            idx = np.flatnonzero(z_epochs >= threshold)
            rank = np.argsort(z_epochs[idx])[::-1]
            res.append((c, z_max[c], idx[rank]))

        if len(res) == 0:
            wx.MessageBox("No rare events were found.", "No Rare Events Found", style=wx.ICON_INFORMATION)
            return

        # format output
        doc = fmtxt.Section("Rare Events")
        doc.add_paragraph(f"Components that disproportionally affect a small number of epochs (z-scored peak-to-peak > {threshold:g}). Epochs are ranked by peak-to-peak.")
        doc.append(fmtxt.linebreak)
        hash_char = {True: fmtxt.FMTextElement('# ', 'font', {'color': 'green'}),
                     False: fmtxt.FMTextElement('# ', 'font', {'color': 'red'})}
        for c, ft, epochs in res:
            doc.append(hash_char[self.doc.accept[c]])
            doc.append(f"{c} ({ft:.1f}):  ")
            doc.append(fmtxt.delim_list((fmtxt.Link(self.doc.epoch_labels[e], f'component:{c} epoch:{e}') for e in epochs)))
            doc.append(fmtxt.linebreak)
        InfoFrame(self, "Rare Events", doc, 500)

    def OnFindBadChannels(self, event):
        nc_before = neighbor_correlation(concatenate(self.doc.epochs_ndvar))
        if self.doc.accept.all():
            nc_after = None
        else:
            epochs = self.doc.as_ndvar(self.doc.apply(self.doc.epochs))
            nc_after = neighbor_correlation(concatenate(epochs))

        # Find ICA components that load on a single channel
        candidates = []
        for i, component_map in enumerate(self.doc.components):
            abs_comp = abs(component_map.x)
            argsort = np.argsort(abs_comp)
            if abs_comp[argsort[-1]] > abs_comp[argsort[-2]] * 3:
                ch_name = self.doc.epochs_ndvar.sensor.names[argsort[-1]]
                # Explained variance
                explained_desc = self.doc.explained_variance(i, format=True)
                explained_variance = max(self.doc.explained_variance(i).values())
                # Loading by epoch
                max_loadings = self.doc.sources[:, i].extrema('time').abs().x
                # Store
                candidates.append([i, ch_name, max_loadings, explained_desc, explained_variance])
        candidates = sorted(candidates, key=itemgetter(-1), reverse=True)

        # format output
        doc = fmtxt.Section("Bad Channels")

        # Neighbor correlation map
        for nc, desc in [[nc_before, 'raw data'], [nc_after, 'cleaned']]:
            if nc is None:
                continue
            figure = matplotlib.figure.Figure(figsize=(4, 3))
            axes = figure.add_axes((0.1, 0.1, .7, 0.8))
            p = plot.Topomap(nc, axes=axes, vmax=1, interpolation='linear')
            p.plot_colorbar(right_of=axes, ticks=3)
            image = fmtxt.Image(f'Neighbor correlation {desc}', 'jpg')
            canvas = FigureCanvasAgg(figure)
            canvas.print_jpeg(image)
            doc.append(image)

        # Candidate components
        for component, ch_name, max_loadings, explained_variance, _ in candidates:
            # plot component map
            figure = matplotlib.figure.Figure(figsize=(1, 1))
            canvas = FigureCanvasAgg(figure)
            axes = figure.add_subplot()
            plot.Topomap(self.doc.components[component], axes=axes, interpolation='linear')
            image = fmtxt.Image(f'#{component}', 'jpg')
            canvas.print_jpeg(image)

            # Text desc
            component_link = fmtxt.Link(f"#{component}", f'component:{component}')
            desc = fmtxt.FMText([ch_name, fmtxt.linebreak, component_link, fmtxt.linebreak, explained_variance])
            table = fmtxt.Table('lll', rules=False)
            doc.add_paragraph(table)

            # Loadings
            binrange = [0, max_loadings.max()]
            figure = matplotlib.figure.Figure(figsize=(2, 1))
            canvas = FigureCanvasAgg(figure)
            axes = figure.add_subplot()
            seaborn.histplot(max_loadings, binrange=binrange, bins=40, ax=axes)
            histogram = fmtxt.Image(f'#{component}', 'jpg')
            canvas.print_jpeg(histogram)

            table.cells(image, desc, histogram)

        InfoFrame(self, "Bad Channels", doc, 500)

    def OnPlotButterfly(self, event):
        self.PlotConditionAverages(self)

    def OnPlotGrandAverage(self, event):
        self.PlotEpochButterfly()

    def OnPlotPSD(self, event):
        self.PlotPSD()

    def OnSetButterflyBaseline(self, event):
        self.butterfly_baseline = event.GetId()

    def PlotConditionAverages(self, parent):
        "Prompt for model and plot condition averages"
        factors = [n for n, v in self.doc.ds.items() if
                   isinstance(v, Factor)]
        if len(factors) == 0:
            wx.MessageBox("The dataset that describes the epochs does not "
                          "contain any Factors that could be used to plot the "
                          "data by condition.", "No Factors in Dataset",
                          style=wx.ICON_ERROR)
            return
        elif len(factors) == 1:
            default = factors[0]
        else:
            default = self.last_model or factors[0]
        msg = "Specify the model (available factors: %s)" % ', '.join(factors)

        plot_model = None
        dlg = wx.TextEntryDialog(parent, msg, "Plot by Condition", default)
        while plot_model is None:
            if dlg.ShowModal() == wx.ID_OK:
                value = dlg.GetValue()
                use = [s.strip() for s in value.replace(':', '%').split('%')]
                invalid = [f for f in use if f not in factors]
                if invalid:
                    wx.MessageBox("The following are not valid factor names: %s"
                                  % (', '.join(invalid)), "Invalid Entry",
                                  wx.ICON_ERROR)
                else:
                    plot_model = '%'.join(use)
            else:
                dlg.Destroy()
                return
        dlg.Destroy()
        self.last_model = value

        ds = self.doc.ds.aggregate(plot_model, drop_bad=True)
        titles = [' '.join(ds[i, f] for f in use) + ' (n=%i)' % ds[i, 'n'] for
                  i in range(ds.n_cases)]
        self._PlotButterfly(ds['epochs'], titles)

    def PlotEpochButterfly(self, i_epoch: int = None):
        if i_epoch is None:
            self._PlotButterfly(self.doc.epochs.average(), "Epochs Average")
        else:
            name = f"Epoch {self.doc.epoch_labels[i_epoch]}"
            self._PlotButterfly(self.doc.epochs[i_epoch], name)

    def _PlotButterfly(self, epoch, title):
        original = self.doc.as_ndvar(epoch)
        clean = self.doc.as_ndvar(self.doc.apply(epoch))
        if self.butterfly_baseline == ID.BASELINE_CUSTOM:
            if original.time.tmin >= 0:
                wx.MessageBox(f"The data displayed does not have a baseline period (tmin={original.time.tmin}). Change the baseline through the Tools menu.", "No Baseline Period", style=wx.ICON_ERROR)
                return
            original -= original.mean(time=(None, 0))
            clean -= clean.mean(time=(None, 0))
        elif self.butterfly_baseline == ID.BASELINE_GLOABL_MEAN:
            original -= self.doc.global_mean
            clean -= self.doc.global_mean

        if original.has_case:
            if isinstance(title, str):
                title = repeat(title, len(original))
            vmax = 1.1 * max(abs(original.min()), original.max())
            for data, title_ in zip(zip(original, clean), title):
                plot.TopoButterfly(data, vmax=vmax, title=title_, axtitle=("Original", "Cleaned"))
        else:
            plot.TopoButterfly([original, clean], title=title, axtitle=("Original", "Cleaned"))

    def PlotPSD(self):
        ds_original = Dataset({'psd': self.doc.as_ndvar(self.doc.epochs).fft().mean('sensor')})
        ds_original[:, 'data'] = 'Source'
        ds_clean = Dataset({'psd': self.doc.as_ndvar(self.doc.apply(self.doc.epochs)).fft().mean('sensor')})
        ds_clean[:, 'data'] = 'Cleaned'
        ds = combine((ds_original, ds_clean))
        colors = {'Source': 'red', 'Cleaned': 'blue'}
        plot.UTSStat('psd', 'data', data=ds, error=np.std, w=8, title="Spectrum (±1 STD)", colors=colors)


class Frame(SharedToolsMenu, FileFrame):
    """GIU for selecting ICA sensor-space components

    Component Selection
    ===================

    * Click on components topographies to select/deselect them.
    * Right-click for a context-menu.

    *Keyboard shortcuts* in addition to the ones in the menu:

    =========== ============================================================
    Key         Effect
    =========== ============================================================
    t           topomap plot of the component under the pointer
    a           array-plot of the source time course of the component
    s           plot sources, starting with the component under the cursor
    f           plot the frequency spectrum for the component under the
                pointer
    b           butterfly plot of grand average (original and cleaned)
    B           butterfly plot of condition averages
    =========== ============================================================
    """
    _doc_name = 'component selection'
    _title = 'Select Components'
    _wildcard = "ICA fiff file (*-ica.fif)|*.fif"

    def __init__(
            self,
            model: Model,
            parent: wx.Frame = None,
            pos: Tuple[int, int] = None,
            size: Tuple[int, int] = None,
    ):
        FileFrame.__init__(self, parent, pos, size, model)
        SharedToolsMenu.__init__(self)
        self.source_frame = None

        # setup layout
        self.ax_size = 200
        figsize = (10, 10)
        self.SetMinSize((400, 400))

        # setup scrolled panel
        panel = ScrolledPanel(self)
        self.panel = panel

        # setup figure canvas
        self.canvas = FigureCanvasPanel(panel, figsize=figsize)
        self.canvas.figure.subplots_adjust(0, 0, 1, 1, 0, 0)
        panel.SetupScrolling(False, scrollToTop=False, scrollIntoView=False)

        # sizer
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.canvas, 0)
        panel.SetSizer(sizer)
        self.canvas_sizer = sizer

        # Toolbar
        tb = self.InitToolbar(can_open=False)
        tb.AddSeparator()
        # buttons
        button = wx.Button(tb, ID.SHOW_SOURCES, "Sources")
        button.Bind(wx.EVT_BUTTON, self.OnShowSources)
        tb.AddControl(button)
        SharedToolsMenu.AddToolbarButtons(self, tb)
        # tail
        tb.AddStretchableSpace()
        self.InitToolbarTail(tb)
        tb.Realize()

        self.CreateStatusBar()

        # Bind Events ---
        self.doc.callbacks.subscribe('case_change', self.CaseChanged)
        self.panel.Bind(wx.EVT_SIZE, self.OnPanelResize)
        self.canvas.mpl_connect('axes_enter_event', self.OnPointerEntersAxes)
        self.canvas.mpl_connect('axes_leave_event', self.OnPointerEntersAxes)
        self.canvas.mpl_connect('button_press_event', self.OnCanvasClick)
        self.canvas.mpl_connect('key_release_event', self.OnCanvasKey)
        # re-Bind right click
        self.canvas.Unbind(wx.EVT_RIGHT_DOWN)
        self.canvas.Unbind(wx.EVT_RIGHT_UP)
        self.canvas.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)

        # Finalize
        self.plot()
        self.UpdateTitle()
        self.canvas.SetFocus()

    def plot(self):
        n = self.doc.ica.n_components_
        fig = self.canvas.figure
        fig.clf()

        panel_w = self.panel.GetSize()[0]
        n_h = max(2, panel_w // self.ax_size)
        n_v = int(ceil(n / n_h))

        # adjust canvas size
        size = (self.ax_size * n_h, self.ax_size * n_v)
        self.canvas_sizer.SetItemMinSize(self.canvas, size)

        # plot
        axes = tuple(fig.add_subplot(n_v, n_h, i) for i in range(1, n + 1))
        # bgs = tuple(ax.patch)
        for i, ax, c, accept in zip(range(n), axes, self.doc.components, self.doc.accept):
            layers = AxisData([DataLayer(c, PlotType.IMAGE)])
            AxTopomap(ax, layers, **TOPO_ARGS)
            ax.text(0.5, 1, "# %i" % i, ha='center', va='top')
            p = Rectangle((0, 0), 1, 1, color=COLOR[accept], zorder=-1)
            ax.add_patch(p)
            ax.i = i
            ax.background = p

        self.axes = axes
        self.n_h = n_h
        self.canvas.store_canvas()
        self.Layout()

    def CaseChanged(self, index):
        "Update the state of the segments on the current page"
        if isinstance(index, INT_TYPES):
            index = [index]
        elif isinstance(index, slice):
            start = index.start or 0
            stop = index.stop or len(self.doc.components)
            index = range(start, stop)
        elif index.dtype.kind == 'b':
            index = np.nonzero(index)[0]

        # update epoch plots
        axes = []
        for idx in index:
            ax = self.axes[idx]
            ax.background.set_color(COLOR[self.doc.accept[idx]])
            axes.append(ax)

        if IS_OSX:
            try:
                self.canvas.redraw(axes)
            except AttributeError:
                self.canvas.draw()
        else:
            self.canvas.draw()  # FIXME: optimize on non-macOS systems

    def FindTopComponent(self, i_epoch: int, only_accepted: bool = False):
        components = self.doc.components
        sources = self.doc.sources.sub(case=i_epoch)
        if only_accepted:
            components = components.sub(component=self.doc.accept)
            sources = sources.sub(component=self.doc.accept)
        comp_power = (components ** 2).sum('sensor')
        source_power = (sources ** 2).sum('time')
        epoch_comp_power = comp_power * source_power
        top_component = epoch_comp_power.argmax()
        self.GoToComponentEpoch(component=top_component)

    def FindTopEpoch(self, i_comp: int):
        source = self.doc.sources.sub(component=i_comp)
        y = source - source.mean()
        y **= 2
        ss = y.sum('time')  # ndvar has epoch as index
        self.GoToComponentEpoch(i_comp, ss.argmax())

    def GoToComponentEpoch(self, component: int = None, epoch: int = None):
        if not self.source_frame:
            self.ShowSources(0)
        self.source_frame.GoToComponentEpoch(component, epoch)

    def MakeToolsMenu(self, menu):
        app = wx.GetApp()
        # show sources
        item = menu.Append(wx.ID_ANY, "Source Viewer", "Open a source time course viewer window")
        app.Bind(wx.EVT_MENU, self.OnShowSources, item)
        # shared menu
        menu.AppendSeparator()
        SharedToolsMenu.MakeToolsMenu(self, menu)

    def OnCanvasClick(self, event):
        "Called by mouse clicks"
        if event.button == 1:
            if event.inaxes:
                self.model.toggle(event.inaxes.i)

    def OnCanvasKey(self, event):
        if not event.inaxes:
            return
        if event.key == 't':
            self.PlotCompTopomap(event.inaxes.i)
        elif event.key == 'a':
            self.PlotCompSourceArray(event.inaxes.i)
        elif event.key == 's':
            self.ShowSources(event.inaxes.i)
        elif event.key == 'f':
            self.PlotCompFFT(event.inaxes.i)
        elif event.key == 'b':
            self.PlotEpochButterfly()
        elif event.key == 'B':
            self.PlotConditionAverages(self)

    def OnFindTopAcceptedComponent(self, event):
        self.FindTopComponent(event.EventObject.i_epoch, only_accepted=True)

    def OnFindTopComponent(self, event):
        self.FindTopComponent(event.EventObject.i_epoch)

    def OnFindTopEpoch(self, event):
        self.FindTopEpoch(event.EventObject.i_comp)

    def OnPanelResize(self, event):
        w, h = event.GetSize()
        n_h = w // self.ax_size
        if n_h >= 2 and n_h != self.n_h:
            self.plot()

    def OnPlotCompFFT(self, event):
        self.PlotCompFFT(event.EventObject.i_comp)

    def OnPlotCompSourceArray(self, event):
        self.PlotCompSourceArray(event.EventObject.i_comp)

    def OnPlotCompTopomap(self, event):
        self.PlotCompTopomap(event.EventObject.i_comp)

    def OnPlotEpoch(self, event):
        self.PlotEpochButterfly(event.EventObject.i_epoch)

    def OnPointerEntersAxes(self, event):
        try:
            sb = self.GetStatusBar()
        except RuntimeError:
            return  # can be called after the window closes (Windows)
        if event.inaxes:
            sb.SetStatusText(f"#{event.inaxes.i} of {len(self.doc.components)} ICA Components")
        else:
            sb.SetStatusText(f"{len(self.doc.components)} ICA Components")

    def OnRankEpochs(self, event):
        i_comp = event.EventObject.i_comp
        source = self.doc.sources.sub(component=i_comp)
        y = source - source.mean()
        y /= y.std()
        y **= 2
        ss = y.sum('time').x  # ndvar has epoch as index

        # sort
        sort = np.argsort(ss)[::-1]

        # doc
        lst = fmtxt.List(f"Epochs SS loading in descending order for component {i_comp}")
        for i in sort:
            link = fmtxt.Link(self.doc.epoch_labels[i], f'component:{i_comp} epoch:{i}')
            lst.add_item(link + f': {ss[i]:.1f}')
        doc = fmtxt.Section(f"#{i_comp} Ranked Epochs", lst)
        InfoFrame(self, f"Component {i_comp} Epoch SS", doc, 200)

    def _context_menu(self, i_comp: int = None, i_epoch: int = None):
        menu = ContextMenu(i_comp, i_epoch)
        if i_comp is not None:
            item = menu.Append(wx.ID_ANY, "Top Epoch")
            self.Bind(wx.EVT_MENU, self.OnFindTopEpoch, item)
            item = menu.Append(wx.ID_ANY, "Rank Epochs")
            self.Bind(wx.EVT_MENU, self.OnRankEpochs, item)
            item = menu.Append(wx.ID_ANY, "Plot Topomap")
            self.Bind(wx.EVT_MENU, self.OnPlotCompTopomap, item)
            item = menu.Append(wx.ID_ANY, "Plot Source Array")
            self.Bind(wx.EVT_MENU, self.OnPlotCompSourceArray, item)
            item = menu.Append(wx.ID_ANY, "Plot Source FFT")
            self.Bind(wx.EVT_MENU, self.OnPlotCompFFT, item)
        if i_comp is not None and i_epoch is not None:
            menu.AppendSeparator()
        if i_epoch is not None:
            item = menu.Append(wx.ID_ANY, "Top Component")
            self.Bind(wx.EVT_MENU, self.OnFindTopComponent, item)
            item = menu.Append(wx.ID_ANY, "Top Accepted Component")
            self.Bind(wx.EVT_MENU, self.OnFindTopAcceptedComponent, item)
            item = menu.Append(wx.ID_ANY, "Plot Epoch")
            self.Bind(wx.EVT_MENU, self.OnPlotEpoch, item)
        return menu

    def OnRightDown(self, event):
        mpl_event = self.canvas._to_matplotlib_event(event)
        if not mpl_event.inaxes:
            return
        menu = self._context_menu(mpl_event.inaxes.i)
        pos = self.panel.CalcScrolledPosition(event.Position)
        self.PopupMenu(menu, pos)
        menu.Destroy()

    def OnShowSources(self, event):
        self.ShowSources(0)

    def OnUpdateUIOpen(self, event):
        event.Enable(False)

    def PlotCompFFT(self, i_comp):
        plot.UTSStat(self.doc.sources.sub(component=i_comp).fft(), error=np.std, w=8, title=f"# {i_comp} Spectrum (±1 STD)", legend=False)

    def PlotCompSourceArray(self, i_comp):
        x = self.doc.sources.sub(component=i_comp)
        dim = Categorial('epoch', self.doc.epoch_labels)
        x = NDVar(x.x, (dim,) + x.dims[1:], x.info, x.name)
        plot.Array(x, w=10, h=10,
                   title='# %i' % i_comp, axtitle=False, interpolation='none')

    def PlotCompTopomap(self, i_comp):
        plot.Topomap(self.doc.components[i_comp], sensorlabels='name', axw=9, title=f'# {i_comp}')

    def ShowSources(self, i_first):
        if self.source_frame:
            self.source_frame.Raise()
        else:
            self.source_frame = SourceFrame(self, i_first)


class SourceFrame(SharedToolsMenu, FileFrameChild):
    """Component source time course display for selecting ICA components.

    * Click on components topographies to select/deselect them.
    * Right-click for a context-menu.

    *Keyboard shortcuts* in addition to the ones in the menu:

    =========== ============================================================
    Key         Effect
    =========== ============================================================
    arrows      scroll through components/epochs
    alt+arrows  scroll to beginning/end
    t           topomap plot of the component under the pointer
    a           array-plot of the source time course of the component under
                the pointer
    f           plot the frequency spectrum for the component under the
                pointer
    b           butterfly plot of the original and cleaned data (of the
                epoch under the pointer, or of the grand average if the
                pointer is over other elements)
    B           Butterfly plot of condition averages
    =========== ============================================================
    """
    _doc_name = 'component selection'
    _title = 'ICA Source Time Course'
    _wildcard = "ICA fiff file (*-ica.fif)|*.fif"

    def __init__(
            self,
            parent: Frame,
            i_first: int,
    ):
        FileFrameChild.__init__(self, parent, None, None, parent.model)
        SharedToolsMenu.__init__(self)

        # prepare canvas
        self.canvas = FigureCanvasPanel(self)
        self.figure = self.canvas.figure
        self.figure.subplots_adjust(0, 0, 1, 1, 0, 0)
        self.figure.set_facecolor('white')

        # attributes
        self.parent = parent
        self.model = parent.model
        self.doc = parent.model.doc
        self.n_comp_actual = self.n_comp = self.config.ReadInt('layout_n_comp', 10)
        self.n_comp_in_ica = len(self.doc.components)
        self.i_first = i_first
        self.n_epochs = self.config.ReadInt('layout_n_epochs', 20)
        self.i_first_epoch = 0
        self.pad_time = 0  # need to pad x-axis when showing fewer epochs than fit on axis)
        self.n_epochs_in_data = len(self.doc.sources)
        self.y_scale = self.config.ReadFloat('y_scale', 10)  # scale factor for y axis
        self._marked_component_i = None
        self._marked_component_h = None
        self._marked_epoch_i = None
        self._marked_epoch_h = None
        self.show_range = True  # show axis with pre/post ICA data range

        # Toolbar
        tb = self.InitToolbar(can_open=False)
        tb.AddSeparator()
        self.up_button = tb.AddTool(wx.ID_UP, "Up", Icon("tango/actions/go-up"))
        self.down_button = tb.AddTool(wx.ID_DOWN, "Down", Icon("tango/actions/go-down"))
        self.back_button = tb.AddTool(wx.ID_BACKWARD, "Back", Icon("tango/actions/go-previous"))
        self.next_button = tb.AddTool(wx.ID_FORWARD, "Next", Icon("tango/actions/go-next"))
        tb.AddSeparator()
        SharedToolsMenu.AddToolbarButtons(self, tb)
        tb.AddStretchableSpace()
        self.InitToolbarTail(tb)
        tb.Realize()

        # event bindings
        self.doc.callbacks.subscribe('case_change', self.CaseChanged)
        self.Bind(wx.EVT_TOOL, self.OnUp, id=wx.ID_UP)
        self.Bind(wx.EVT_TOOL, self.OnDown, id=wx.ID_DOWN)
        self.Bind(wx.EVT_TOOL, self.OnBackward, id=wx.ID_BACKWARD)
        self.Bind(wx.EVT_TOOL, self.OnForward, id=wx.ID_FORWARD)
        self.canvas.mpl_connect('key_release_event', self.OnCanvasKey)
        # re-Bind mouse click
        self.canvas.Unbind(wx.EVT_LEFT_DOWN)
        self.canvas.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.canvas.Unbind(wx.EVT_RIGHT_DOWN)
        self.canvas.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)

        self._plot()
        self.UpdateTitle()
        self.Show()

    def _get_source_data(self):
        "Return ``(source_data, epoch-labels)`` tuple for current page"
        n_comp = self.n_comp
        n_comp_actual = self.n_comp_actual
        epoch_index = slice(self.i_first_epoch, self.i_first_epoch + self.n_epochs)
        data = self.doc.sources.sub(case=epoch_index, component=slice(self.i_first, self.i_first + n_comp))
        y = data.get_data(('component', 'case', 'time')).reshape((n_comp_actual, -1))
        if y.base is not None and data.x.base is not None:
            y = y.copy()
        start = n_comp - 1 + self.show_range
        stop = -1 + (n_comp - n_comp_actual) + self.show_range
        y += np.arange(start * self.y_scale, stop * self.y_scale, -self.y_scale)[:, None]
        # pad epoch labels for x-axis
        epoch_labels = self.doc.epoch_labels[epoch_index]
        if len(epoch_labels) < self.n_epochs:
            epoch_labels += ('',) * (self.n_epochs - len(epoch_labels))
        return y, epoch_labels

    def _pad(self, y):
        "Pad time-axis when data contains fewer epochs than the x-axis"
        if self.pad_time:
            return np.pad(y, (0, self.pad_time), 'constant')
        else:
            return y

    def _get_raw_range(self):
        epoch_index = slice(self.i_first_epoch, self.i_first_epoch + self.n_epochs)
        y_min = self._pad(self.doc.pre_ica_min[epoch_index].x.ravel())
        y_max = self._pad(self.doc.pre_ica_max[epoch_index].x.ravel())
        return y_min, y_max

    def _get_clean_range(self):
        epoch_index = slice(self.i_first_epoch, self.i_first_epoch + self.n_epochs)
        epochs = self.doc.epochs[epoch_index]
        y_clean = self.doc.as_ndvar(self.doc.apply(epochs))
        y_min = y_clean.min('sensor').x.ravel()
        y_max = y_clean.max('sensor').x.ravel()
        y_min /= self.doc.pre_ica_range_scale
        y_max /= self.doc.pre_ica_range_scale
        return self._pad(y_min), self._pad(y_max)

    def _plot(self):
        # partition figure
        self.figure.clf()
        figheight = self.figure.get_figheight()
        n_comp = self.n_comp
        # make sure there are no empty lines
        if self.i_first and self.n_comp_in_ica - self.i_first < n_comp:
            self.i_first = max(0, self.n_comp_in_ica - n_comp)
        # further layout-relevant properties
        n_comp_actual = min(self.n_comp_in_ica - self.i_first, n_comp)
        self.n_comp_actual = n_comp_actual
        elen = len(self.doc.sources.time)

        # layout
        n_rows = n_comp + self.show_range
        axheight = 1 / (n_rows + 0.5)  # 0.5 = bottom space for epoch labels

        # topomaps
        ax_size_in = axheight * figheight
        axwidth = ax_size_in / self.figure.get_figwidth()
        left = axwidth / 2
        self.topo_plots = []
        self.topo_labels = []
        for i in range(n_comp_actual):
            i_comp = self.i_first + i
            ax = self.figure.add_axes((left, 1 - (i + 1) * axheight, axwidth, axheight))
            layers = AxisData([DataLayer(self.doc.components[i_comp], PlotType.IMAGE)])
            p = AxTopomap(ax, layers, **TOPO_ARGS)
            text = ax.text(0, 0.5, "# %i" % i_comp, va='center', ha='right', color='k')
            ax.i = i
            ax.i_comp = i_comp
            self.topo_plots.append(p)
            self.topo_labels.append(text)

        # source time course data
        y, xtick_labels = self._get_source_data()

        # axes
        left = 1.5 * axwidth
        bottom = 1 - n_rows * axheight
        xticks = np.arange(elen / 2, elen * self.n_epochs, elen)
        ax = self.figure.add_axes((left, bottom, 1 - left, 1 - bottom), frameon=False, yticks=(), xticks=xticks, xticklabels=xtick_labels)
        ax.tick_params(bottom=False)
        ax.i = -1
        ax.i_comp = None

        # store canvas before plotting lines
        self.canvas.draw()

        # plot epochs
        self.lines = ax.plot(y.T, color=LINE_COLOR[True], clip_on=False)
        # line color
        reject_color = LINE_COLOR[False]
        for i in range(n_comp_actual):
            if not self.doc.accept[i + self.i_first]:
                self.lines[i].set_color(reject_color)
        # data pre/post range
        if self.show_range:
            pre_color = UNAMBIGUOUS_COLORS['orange']
            post_color = UNAMBIGUOUS_COLORS['bluish green']
            ax.text(-10, 0.1, 'Range: Raw', va='bottom', ha='right', color=pre_color)
            ax.text(-10, -0.1, 'Cleaned', va='top', ha='right', color=post_color)
            # raw
            ys_raw = self._get_raw_range()
            self.y_range_pre_lines = [ax.plot(yi, color=pre_color, clip_on=False)[0] for yi in ys_raw]
            # cleaned
            ys_clean = self._get_clean_range()
            self.y_range_post_lines = [ax.plot(yi, color=post_color, clip_on=False)[0] for yi in ys_clean]
        # axes limits
        self.ax_tc_ylim = (-0.5 * self.y_scale, (n_rows - 0.5) * self.y_scale)
        ax.set_ylim(self.ax_tc_ylim)
        ax.set_xlim((0, y.shape[1]))
        # epoch demarcation
        for x in range(elen, elen * self.n_epochs, elen):
            ax.axvline(x, ls='--', c='k')

        self.ax_tc = ax
        self.canvas.draw()

    def _plot_update_raw_range(self):
        y_min, y_max = self._get_raw_range()
        for line, data in zip(self.y_range_pre_lines, (y_min, y_max)):
            line.set_ydata(data)

    def _plot_update_clean_range(self):
        y_min, y_max = self._get_clean_range()
        for line, data in zip(self.y_range_post_lines, (y_min, y_max)):
            line.set_ydata(data)

    def _event_i_comp(self, event):
        if event.inaxes:
            if event.inaxes.i_comp is None:
                i_in_axes = ceil(event.ydata / self.y_scale + 0.5)
                if i_in_axes == 1 and self.show_range:
                    return
                i_comp = int(self.i_first + self.n_comp + self.show_range - i_in_axes)
                if i_comp < self.n_comp_in_ica:
                    return i_comp
            else:
                return event.inaxes.i_comp

    def _event_i_epoch(self, event):
        if event.inaxes is not None and event.inaxes.i_comp is None:
            i_epoch = self.i_first_epoch + int(event.xdata // len(self.doc.sources.time))
            if 0 <= i_epoch < len(self.doc.epochs):
                return i_epoch

    def CanBackward(self):
        return self.i_first_epoch > 0

    def CanDown(self):
        return self.i_first + self.n_comp < self.n_comp_in_ica

    def CanForward(self):
        return self.i_first_epoch + self.n_epochs < self.n_epochs_in_data

    def CanUp(self):
        return self.i_first > 0

    def CaseChanged(self, index):
        "Update the states of the segments on the current page"
        if isinstance(index, int):
            index = [index]
        elif isinstance(index, slice):
            start = index.start or 0
            stop = index.stop or self.doc.n_epochs
            index = range(start, stop)
        elif index.dtype.kind == 'b':
            index = np.nonzero(index)[0]

        # filter to visible epochs
        i_last = self.i_first + self.n_comp_actual
        index = [i_comp for i_comp in index if self.i_first <= i_comp <= i_last]
        # update epoch plots
        if index:
            for i_comp in index:
                self.lines[i_comp - self.i_first].set_color(LINE_COLOR[self.doc.accept[i_comp]])
            self._plot_update_clean_range()
            self.canvas.draw()

    def GoToComponentEpoch(self, component: int = None, epoch: int = None):
        if component is not None:
            self._marked_component_i = component
            self.SetFirstComponent(component // self.n_comp * self.n_comp)
        if epoch is not None:
            self._marked_epoch_i = epoch
            self.SetFirstEpoch(epoch // self.n_epochs * self.n_epochs)
        self.Raise()

    def OnBackward(self, event):
        "Turn the page backward"
        self.SetFirstEpoch(self.i_first_epoch - self.n_epochs)

    def OnCanvasKey(self, event):
        if event.key is None:
            return
        elif event.key == 'alt+down':
            self.SetFirstComponent(self.n_comp_in_ica - self.n_comp)
        elif event.key == 'down':
            if self.CanDown():
                self.OnDown(None)
        elif event.key == 'alt+up':
            self.SetFirstComponent(0)
        elif event.key == 'up':
            if self.CanUp():
                self.OnUp(None)
        elif event.key == 'alt+right':
            self.SetFirstEpoch(((self.n_epochs_in_data - 1) // self.n_epochs) * self.n_epochs)
        elif event.key == 'right':
            if self.CanForward():
                self.OnForward(None)
        elif event.key == 'alt+left':
            self.SetFirstEpoch(0)
        elif event.key == 'left':
            if self.CanBackward():
                self.OnBackward(None)
        elif event.key == 'B':
            self.parent.PlotConditionAverages(self)
        elif event.key == 'b':
            self.PlotEpochButterfly(self._event_i_epoch(event))
        elif not event.inaxes:
            return
        # component-specific plots
        i_comp = self._event_i_comp(event)
        if i_comp is None:  # source time course axes
            return
        elif event.key in 'tT':
            self.parent.PlotCompTopomap(i_comp)
        elif event.key == 'a':
            self.parent.PlotCompSourceArray(i_comp)
        elif event.key == 'f':
            self.parent.PlotCompFFT(i_comp)

    def OnClose(self, event):
        if super(SourceFrame, self).OnClose(event):
            self.doc.callbacks.remove('case_change', self.CaseChanged)
            self.config.WriteInt('layout_n_comp', self.n_comp)
            self.config.WriteInt('layout_n_epochs', self.n_epochs)
            self.config.WriteFloat('y_scale', self.y_scale)
            self.config.Flush()

    def OnDown(self, event):
        "Turn the page backward"
        self.SetFirstComponent(self.i_first + self.n_comp)

    def OnForward(self, event):
        "Turn the page forward"
        self.SetFirstEpoch(self.i_first_epoch + self.n_epochs)

    def OnLeftDown(self, event):
        "Called by mouse clicks"
        mpl_event = self.canvas._to_matplotlib_event(event)
        i_comp = self._event_i_comp(mpl_event)
        if i_comp is None:
            return
        self.model.toggle(i_comp)

    def OnRightDown(self, event):
        mpl_event = self.canvas._to_matplotlib_event(event)
        i_comp = self._event_i_comp(mpl_event)
        i_epoch = self._event_i_epoch(mpl_event)
        if i_comp is None and i_epoch is None:
            return
        menu = self.parent._context_menu(i_comp, i_epoch)
        self.PopupMenu(menu, event.Position)
        menu.Destroy()

    def OnSetLayout(self, event):
        caption = "Set ICA Source Layout"
        msg = "Number of components and epochs (e.g., '10 20')"
        default = '%i %i' % (self.n_comp, self.n_epochs)
        dlg = wx.TextEntryDialog(self, msg, caption, default)
        while True:
            if dlg.ShowModal() == wx.ID_OK:
                value = dlg.GetValue()
                try:
                    n_comp, n_epochs = map(int, value.split())
                except Exception:
                    wx.MessageBox("Invalid entry: %r. Need two integers \n"
                                  "(e.g., '10 20').", "Invalid Entry",
                                  wx.OK | wx.ICON_ERROR)
                else:
                    dlg.Destroy()
                    break
            else:
                dlg.Destroy()
                return

        self.n_comp = n_comp
        self.n_epochs = n_epochs
        self._plot()

    def OnSetVLim(self, event):
        dlg = wx.TextEntryDialog(self, "Y-axis scale:", "Y-Axis Scale",
                                 "%g" % (10. / self.y_scale,))
        value = None
        while True:
            if dlg.ShowModal() != wx.ID_OK:
                break
            error = None
            try:
                value = float(dlg.GetValue())
                if value <= 0:
                    error = f"{value}: must be > 0"
            except Exception as exception:
                error = str(exception)

            if error:
                msg = wx.MessageDialog(self, error, "Invalid Entry", wx.OK | wx.ICON_ERROR)
                msg.ShowModal()
                msg.Destroy()
            else:
                break
        dlg.Destroy()
        if value is not None:
            self.y_scale = 10. / value
            # redraw
            self.SetFirstEpoch(self.i_first_epoch)

    def OnUp(self, event):
        "Turn the page backward"
        self.SetFirstComponent(self.i_first - self.n_comp)

    def OnUpdateUIBackward(self, event):
        event.Enable(self.CanBackward())

    def OnUpdateUIDown(self, event):
        event.Enable(self.CanDown())

    def OnUpdateUIForward(self, event):
        event.Enable(self.CanForward())

    def OnUpdateUISetLayout(self, event):
        event.Enable(True)

    def OnUpdateUISetVLim(self, event):
        event.Enable(True)

    def OnUpdateUIUp(self, event):
        event.Enable(self.CanUp())

    def SetFirstComponent(self, i_first):
        if i_first < 0:
            i_first = 0
        elif i_first >= self.n_comp_in_ica:
            i_first = self.n_comp_in_ica - 1
        n_rows = self.n_comp + self.show_range

        # marked component
        if self._marked_component_h is not None:
            self._marked_component_h.remove()
            self._marked_component_h = None
        if self._marked_component_i is not None:
            i_from_top = self._marked_component_i - i_first
            i_from_bottom = n_rows - 1 - i_from_top
            if 0 <= i_from_bottom < n_rows:
                bottom = (i_from_bottom - 0.5) * self.y_scale
                self._marked_component_h = self.ax_tc.axhspan(bottom, bottom + self.y_scale, edgecolor='yellow', facecolor='yellow')

        n_comp_actual = min(self.n_comp_in_ica - i_first, self.n_comp)
        for i in range(n_comp_actual):
            p = self.topo_plots[i]
            i_comp = i_first + i
            p.set_data([self.doc.components[i_comp]], True)
            p.ax.i_comp = i_comp
            self.topo_labels[i].set_text("# %i" % i_comp)
            self.lines[i].set_color(LINE_COLOR[self.doc.accept[i_comp]])

        if n_comp_actual < self.n_comp:
            empty_data = self.doc.components[0].copy()
            empty_data.x.fill(0)
            for i in range(n_comp_actual, self.n_comp):
                p = self.topo_plots[i]
                p.set_data([empty_data])
                p.ax.i_comp = -1
                self.topo_labels[i].set_text("")
                self.lines[i].set_color('white')

        self.i_first = i_first
        self.n_comp_actual = n_comp_actual
        self.SetFirstEpoch(self.i_first_epoch)

    def SetFirstEpoch(self, i_first_epoch):
        self.i_first_epoch = i_first_epoch

        # marked epoch
        if self._marked_epoch_h is not None:
            self._marked_epoch_h.remove()
            self._marked_epoch_h = None
        if self._marked_epoch_i is not None:
            i = self._marked_epoch_i - i_first_epoch
            if 0 <= i < self.n_epochs:
                elen = len(self.doc.sources.time)
                bottom = -0.5 * self.y_scale
                height = (self.n_comp + self.show_range) * self.y_scale
                self._marked_epoch_h = Rectangle((i * elen, bottom), elen, height, edgecolor='yellow', facecolor='yellow')
                self.ax_tc.add_patch(self._marked_epoch_h)

        # update data
        y, tick_labels = self._get_source_data()
        if i_first_epoch + self.n_epochs > self.n_epochs_in_data:
            elen = len(self.doc.sources.time)
            n_missing = self.i_first_epoch + self.n_epochs - self.n_epochs_in_data
            pad_time = elen * n_missing
        else:
            pad_time = 0
        self.pad_time = pad_time

        if self.n_comp_actual < self.n_comp:
            pad_comp = self.n_comp - self.n_comp_actual
        else:
            pad_comp = 0

        if pad_time or pad_comp:
            y = np.pad(y, ((0, pad_comp), (0, pad_time)), 'constant')

        for line, data in zip(self.lines, y):
            line.set_ydata(data)

        if self.show_range:
            self._plot_update_raw_range()
            self._plot_update_clean_range()

        self.ax_tc.set_xticklabels(tick_labels)
        self.ax_tc.set_ylim(self.ax_tc_ylim)
        self.canvas.draw()


class FindNoisyEpochsDialog(EelbrainDialog):
    _default_thresholds = {'µV': 100, 'fT': 1000}

    def __init__(self, parent, unit, **kwargs):
        self.unit = unit
        super(FindNoisyEpochsDialog, self).__init__(parent, wx.ID_ANY, "Find Bad Epochs", **kwargs)
        config = parent.config
        threshold = config.ReadFloat(f"FindNoisyEpochsDialog/threshold_{unit}", self._default_threshold())
        apply_rejection = config.ReadBool("FindNoisyEpochsDialog/apply_rejection", True)
        sort_by_component = config.ReadBool("FindNoisyEpochsDialog/sort_by_component", True)
        max_ch_ratio = config.Read("FindNoisyEpochsDialog/max_ch_ratio", '')

        sizer = wx.BoxSizer(wx.VERTICAL)

        # Threshold
        h_sizer = wx.BoxSizer(wx.HORIZONTAL)
        h_sizer.Add(wx.StaticText(self, label="Threshold for bad epochs: "))
        validator = REValidator(POS_FLOAT_PATTERN, "Invalid entry: {value}. Please specify a number > 0.", False)
        self.threshold = ctrl = wx.TextCtrl(self, value=f'{threshold:g}', validator=validator, style=wx.TE_RIGHT)
        ctrl.SetHelpText("Find epochs in which the signal exceeds this value at any sensor")
        ctrl.SelectAll()
        h_sizer.Add(ctrl)
        h_sizer.Add(wx.StaticText(self, label=unit))
        sizer.Add(h_sizer)

        # Apply rejection before finding noisy epochs
        self.apply_rejection = ctrl = wx.CheckBox(self, label="Apply ICA rejection")
        ctrl.SetValue(apply_rejection)
        sizer.Add(ctrl)

        # Sort noisy epochs by component
        self.sort_by_component = ctrl = wx.CheckBox(self, label="Sort by ICA component")
        ctrl.SetValue(sort_by_component)
        sizer.Add(ctrl)

        # Filter by ch 1/2 ratio
        h_sizer = wx.BoxSizer(wx.HORIZONTAL)
        h_sizer.Add(wx.StaticText(self, label="Max channel ratio filter: "))
        validator = REValidator(FLOAT_PATTERN, "Invalid entry: {value}. Please specify a number ≥ 0.", True)
        self.max_ch_ratio = ctrl = wx.TextCtrl(self, value=max_ch_ratio, validator=validator, style=wx.TE_RIGHT)
        ctrl.SetHelpText("Filter components that are due to bad channels through the first / second channel ratio")
        ctrl.SelectAll()
        h_sizer.Add(ctrl)
        sizer.Add(h_sizer)

        # default button
        btn = wx.Button(self, wx.ID_DEFAULT, "Default Settings")
        sizer.Add(btn, border=2)
        btn.Bind(wx.EVT_BUTTON, self.OnSetDefault)

        # buttons
        button_sizer = wx.StdDialogButtonSizer()
        # ok
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        button_sizer.AddButton(btn)
        # cancel
        btn = wx.Button(self, wx.ID_CANCEL)
        button_sizer.AddButton(btn)
        # finalize
        button_sizer.Realize()
        sizer.Add(button_sizer)

        self.SetSizer(sizer)
        sizer.Fit(self)

    def _default_threshold(self):
        return self._default_thresholds.get(self.unit, 1)

    def OnSetDefault(self, event):
        self.threshold.SetValue(f'{self._default_threshold()}')

    def StoreConfig(self):
        config = self.Parent.config
        config.WriteFloat(f"FindNoisyEpochsDialog/threshold_{self.unit}", float(self.threshold.GetValue()))
        config.WriteBool("FindNoisyEpochsDialog/apply_rejection", self.apply_rejection.GetValue())
        config.WriteBool("FindNoisyEpochsDialog/sort_by_component", self.sort_by_component.GetValue())
        config.Write("FindNoisyEpochsDialog/max_ch_ratio", self.max_ch_ratio.GetValue())
        config.Flush()


class FindRareEventsDialog(EelbrainDialog):
    def __init__(self, parent, *args, **kwargs):
        super(FindRareEventsDialog, self).__init__(parent, wx.ID_ANY, "Find Rare Events", *args, **kwargs)
        config = parent.config
        threshold = config.ReadFloat("FindRareEvents/threshold", 2.)

        sizer = wx.BoxSizer(wx.VERTICAL)

        # Threshold
        sizer.Add(wx.StaticText(self, label="Threshold for rare epochs\n(z-scored peak-to-peak value):"))
        validator = REValidator(POS_FLOAT_PATTERN, "Invalid entry: {value}. Please specify a number > 0.", False)
        ctrl = wx.TextCtrl(self, value=str(threshold), validator=validator)
        ctrl.SetHelpText("Epochs whose z-scored peak-to-peak value exceeds  this value are considered rare")
        ctrl.SelectAll()
        sizer.Add(ctrl)
        self.threshold = ctrl

        # default button
        btn = wx.Button(self, wx.ID_DEFAULT, "Default Settings")
        sizer.Add(btn, border=2)
        btn.Bind(wx.EVT_BUTTON, self.OnSetDefault)

        # buttons
        button_sizer = wx.StdDialogButtonSizer()
        # ok
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        button_sizer.AddButton(btn)
        # cancel
        btn = wx.Button(self, wx.ID_CANCEL)
        button_sizer.AddButton(btn)
        # finalize
        button_sizer.Realize()
        sizer.Add(button_sizer)

        self.SetSizer(sizer)
        sizer.Fit(self)

    def OnSetDefault(self, event):
        self.threshold.SetValue('2')

    def StoreConfig(self):
        config = self.Parent.config
        config.WriteFloat("FindRareEvents/threshold", float(self.threshold.GetValue()))
        config.Flush()


class InfoFrame(HTMLFrame):

    def __init__(
            self,
            parent: Union[wx.Window, SharedToolsMenu],
            title: str,
            doc: fmtxt.FMTextElement,
            w: int,
            h: int = -1,
    ):
        pos, size = self.find_pos(w, h)
        style = wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX | wx.RESIZE_BORDER | wx.CAPTION | wx.CLOSE_BOX | wx.FRAME_FLOAT_ON_PARENT | wx.FRAME_TOOL_WINDOW
        html_doc = fmtxt.make_html_doc(doc)
        HTMLFrame.__init__(self, parent, title, html_doc, pos=pos, size=size, style=style)

    @staticmethod
    def find_pos(w: int, h: int):
        display_w, display_h = wx.DisplaySize()
        h_max = display_h - 44
        h = h_max if h <= 0 else min(h, h_max)
        pos = (display_w - w, int(round((display_h - h) / 2)))
        return pos, (w, h)

    def OpenURL(self, url):
        component = epoch = None
        for part in url.split():
            m = re.match(r'^epoch:(\d+)$', part)
            if m:
                epoch = int(m.group(1))
                continue
            m = re.match(r'^component:(\d+)$', part)
            if m:
                component = int(m.group(1))
                continue
            raise ValueError(f"url={url!r}")
        self.Parent.GoToComponentEpoch(component, epoch)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from dataclasses import dataclass
from enum import Enum
from functools import partial, reduce
from itertools import product
from numbers import Number
import operator
from typing import Any, Literal, Sequence, Union

import matplotlib
import mne
from nibabel.freesurfer import read_annot
import numpy as np

from .._data_obj import NDVar, SourceSpace, UTS
from .._types import PathArg
from .._stats.testnd import NDTest, NDDifferenceTest, MultiEffectNDTest
from .._utils import deprecated
from ..fmtxt import im_table
from .._text import ms, ms_window
from ._base import EelFigure, ImLayout, ColorBarMixin, brain_data, butterfly_data, use_inline_backend
from ._color_luts import dspm_lut, p_lut
from ._colors import ColorList
from ._utsnd import Butterfly


def annot(annot, subject='fsaverage', surf='smoothwm', borders=False, alpha=0.7,
          hemi=None, views=('lat', 'med'), w=None, h=None, axw=None, axh=None,
          foreground=None, background=None, parallel=True, cortex='classic',
          title=None, subjects_dir=None, name=None):
    """Plot the parcellation in an annotation file

    Parameters
    ----------
    annot : str
        Name of the annotation (e.g., "PALS_B12_LOBES").
    subject : str
        Name of the subject (default 'fsaverage').
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    borders : bool | int
        Show only label borders (PySurfer Brain.add_annotation() argument).
    alpha : scalar
        Alpha of the annotation (1=opaque, 0=transparent, default 0.7).
    hemi : 'lh' | 'rh' | 'both' | 'split'
        Which hemispheres to plot (default includes hemisphere with more than one
        label in the annot file).
    views : str | sequence of str
        One or several views to show in the figure. The options are:
        ``'lateral', 'medial', 'ventral', 'dorsal', 'rostral', 'parietal',
        'frontal', 'caudal'``.
    w, h, axw, axh : scalar
        Layout parameters (figure width/height, subplot width/height).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    title : str
        title for the window (default is the parcellation name).
    subjects_dir : None | str
        Override the default subjects_dir.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.

    Returns
    -------
    brain : surfer.Brain
        PySurfer Brain instance.

    Notes
    -----
    The ``Brain`` object that is returned has a
    :meth:`~plot._brain_fixes.plot_legend` method to plot the color legend.

    See Also
    --------
    eelbrain.plot.brain.annot_legend : plot a corresponding legend without the brain
    """
    if hemi is None:
        annot_lh = mne.read_labels_from_annot(subject, annot, 'lh',
                                              subjects_dir=subjects_dir)
        use_lh = len(annot_lh) > 1
        annot_rh = mne.read_labels_from_annot(subject, annot, 'rh',
                                              subjects_dir=subjects_dir)
        use_rh = len(annot_rh) > 1
        if use_lh and use_rh:
            hemi = 'split'
        elif use_lh:
            hemi = 'lh'
        elif use_rh:
            hemi = 'rh'
        else:
            raise ValueError("Neither hemisphere contains more than one label")

    if title is None:
        title = '%s - %s' % (subject, annot)

    from ._brain_object import Brain
    brain = Brain(subject, hemi, surf, title, cortex,
                  views=views, w=w, h=h, axw=axw, axh=axh,
                  foreground=foreground, background=background,
                  subjects_dir=subjects_dir, name=name)

    brain._set_annot(annot, borders, alpha)
    if parallel:
        brain.set_parallel_view(scale=True)
    return brain


def annot_legend(lh, rh, *args, **kwargs):
    """Plot a legend for a freesurfer parcellation

    Parameters
    ----------
    lh : str
        Path to the lh annot-file.
    rh : str
        Path to the rh annot-file.
    labels : dict (optional)
        Alternative (text) label for (brain) labels.
    h : 'auto' | scalar
        Height of the figure in inches. If 'auto' (default), the height is
        automatically increased to fit all labels.

    Returns
    -------
    legend : :class:`~eelbrain.plot.ColorList`
        Figure with legend for the parcellation.

    Notes
    -----
    Instead of :func:`~eelbrain.plot.brain.annot_legend` it is usually
    easier to use::

    >>> brain = plot.brain.annoot(annot, ...)
    >>> legend = brain.plot_legend()

    See Also
    --------
    eelbrain.plot.brain.annot : plot the parcellation on a brain model
    """
    _, lh_colors, lh_names = read_annot(lh)
    _, rh_colors, rh_names = read_annot(rh)
    lh_names = [name.decode() for name in lh_names]
    rh_names = [name.decode() for name in rh_names]
    lh_colors = dict(zip(lh_names, lh_colors[:, :4] / 255.))
    rh_colors = dict(zip(rh_names, rh_colors[:, :4] / 255.))
    names = set(lh_names)
    names.update(rh_names)
    colors = {}
    seq = []  # sequential order in legend
    seq_lh = []
    seq_rh = []
    for name in names:
        if name in lh_colors and name in rh_colors:
            if np.array_equal(lh_colors[name], rh_colors[name]):
                colors[name] = lh_colors[name]
                seq.append(name)
            else:
                colors[name + '-lh'] = lh_colors[name]
                colors[name + '-rh'] = rh_colors[name]
                seq_lh.append(name + '-lh')
                seq_rh.append(name + '-rh')
        elif name in lh_colors:
            colors[name + '-lh'] = lh_colors[name]
            seq_lh.append(name + '-lh')
        else:
            colors[name + '-rh'] = rh_colors[name]
            seq_rh.append(name + '-rh')
    return ColorList(colors, seq + seq_lh + seq_rh, *args, **kwargs)


def _plot(data, *args, **kwargs):
    "Plot depending on source space kind"
    if data.source.kind == 'vol':
        return _voxel_brain(data, *args, **kwargs)
    else:
        return brain(data, *args, **kwargs)


def dspm(src, fmin=13, fmax=22, fmid=None, *args, **kwargs):
    """
    Plot a source estimate with coloring for dSPM values (bipolar).

    Parameters
    ----------
    src : NDVar, dims = ([case,] source, [time])
        NDVar with SourceSpace dimension. If stc contains a case dimension,
        the average across cases is taken.
    fmin, fmax : scalar >= 0
        Start- and end-point for the color gradient for positive values. The
        gradient for negative values goes from -fmin to -fmax. Values between
        -fmin and fmin are transparent.
    fmid : None | scalar
        Midpoint for the color gradient. If fmid is None (default) it is set
        half way between fmin and fmax.
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : str | sequence of str
        One or several views to show in the figure. The options are:
        ``'lateral', 'medial', 'ventral', 'dorsal', 'rostral', 'parietal',
        'frontal', 'caudal'``.
    hemi : 'lh' | 'rh' | 'both' | 'split'
        Which hemispheres to plot (default based on data).
    colorbar : bool
        Add a colorbar to the figure (use ``.plot_colorbar()`` to plot a
        colorbar separately).
    time_label : str
        Label to show time point. Use ``'ms'`` or ``'s'`` to display time in
        milliseconds or in seconds, or supply a custom format string to format
        time values (in seconds; default is ``'ms'``).
    w, h, axw, axh : scalar
        Layout parameters (figure width/height, subplot width/height).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    title : str
        title for the window (default is the subject name).
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``src``. Can be matplotlib color, including
        alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white).
    subjects_dir : None | str
        Override the subjects_dir associated with the source space dimension.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.

    Returns
    -------
    brain : surfer.Brain
        PySurfer Brain instance containing the plot.
    """
    if fmid is None:
        fmid = (fmax + fmin) / 2
    lut = dspm_lut(fmin, fmid, fmax)
    return _plot(src, lut, -fmax, fmax, *args, **kwargs)


def p_map(p_map, param_map=None, p0=0.05, p1=0.01, p0alpha=0.5, *args,
          **kwargs):
    """Plot a map of p-values in source space.

    Parameters
    ----------
    p_map : NDVar | NDTest
        Map of p values, or test result.
    param_map : NDVar
        Statistical parameter covering the same data points as p_map. Only the
        sign is used, for incorporating the directionality of the effect into
        the plot.
    p0 : scalar
        Highest p-value that is visible.
    p1 : scalar
        P-value where the colormap changes from ramping alpha to ramping color.
    p0alpha : 1 >= float >= 0
        Alpha at ``p0``. Set to 0 for a smooth transition, or a larger value to
        clearly delineate significant regions (default 0.5).
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : str | sequence of str
        One or several views to show in the figure. The options are:
        ``'lateral', 'medial', 'ventral', 'dorsal', 'rostral', 'parietal',
        'frontal', 'caudal'``.
    hemi : 'lh' | 'rh' | 'both' | 'split'
        Which hemispheres to plot (default based on data).
    colorbar : bool
        Add a colorbar to the figure (use ``.plot_colorbar()`` to plot a
        colorbar separately).
    time_label : str
        Label to show time point. Use ``'ms'`` or ``'s'`` to display time in
        milliseconds or in seconds, or supply a custom format string to format
        time values (in seconds; default is ``'ms'``).
    w, h, axw, axh : scalar
        Layout parameters (figure width/height, subplot width/height).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    title : str
        title for the window (default is the subject name).
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``p_map``. Can be matplotlib color,
        including alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white).
    subjects_dir : None | str
        Override the subjects_dir associated with the source space dimension.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.

    Returns
    -------
    brain : surfer.Brain
        PySurfer Brain instance containing the plot.

    Notes
    -----
    In order to make economical use of the color lookup table, p-values are
    remapped for display. P-values larger than ``p0`` are mapped to 0, p-values
    for positive effects to ``[step, p0 + step]`` and p-values
    for negative effects to ``[-step, -(p0 + step)]``.

    Due to this, in order to plot a colorbar only including positive
    differences yse::

    >>> brain.plot_colorbar(clipmin=0)

    and to include only negative effects::

    >>> brain.plot_colorbar(clipmax=0)
    """
    from .._stats.testnd import NDTest, MultiEffectNDTest

    if isinstance(p_map, NDTest):
        if isinstance(p_map, MultiEffectNDTest):
            raise NotImplementedError(f"plot.brain.p_map for {p_map.__class__.__name__}")
        source = p_map.p.source
    else:
        source = p_map.source
    p = brain(source, None, None, None, *args, **kwargs)
    p.add_ndvar_p_map(p_map, param_map, p0, p1, p0alpha)
    return p


def cluster(cluster, vmax=None, *args, **kwargs):
    """Plot a spatio-temporal cluster

    Plots a cluster with the assumption that all non-zero data should be
    visible, while areas that not part of the cluster are 0.

    Parameters
    ----------
    cluster : NDVar
        The cluster.
    vmax : scalar != 0
        Maximum value in the colormap. Default is the maximum value in the
        cluster.
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : str | sequence of str
        One or several views to show in the figure. The options are:
        ``'lateral', 'medial', 'ventral', 'dorsal', 'rostral', 'parietal',
        'frontal', 'caudal'``.
    hemi : 'lh' | 'rh' | 'both' | 'split'
        Which hemispheres to plot (default based on data).
    colorbar : bool
        Add a colorbar to the figure (use ``.plot_colorbar()`` to plot a
        colorbar separately).
    time_label : str
        Label to show time point. Use ``'ms'`` or ``'s'`` to display time in
        milliseconds or in seconds, or supply a custom format string to format
        time values (in seconds; default is ``'ms'``).
    w, h, axw, axh : scalar
        Layout parameters (figure width/height, subplot width/height).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    title : str
        title for the window (default is the subject name).
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``cluster``. Can be matplotlib color,
        including alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white).
    subjects_dir : None | str
        Override the subjects_dir associated with the source space dimension.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.

    Returns
    -------
    brain : surfer.Brain
        PySurfer Brain instance containing the plot.
    """
    if vmax is None:
        vmax = max(cluster.x.max(), -cluster.x.min())
        if vmax == 0:
            raise ValueError("The cluster's data is all zeros")
    elif vmax == 0:
        raise ValueError("vmax can't be 0")
    elif vmax < 0:
        vmax = -vmax

    lut = dspm_lut(0, vmax / 10, vmax)
    return _plot(cluster, lut, -vmax, vmax, *args, **kwargs)


def brain(src, cmap=None, vmin=None, vmax=None, surf='inflated',
          views='lateral', hemi=None, colorbar=False, time_label='ms',
          w=None, h=None, axw=None, axh=None, foreground=None, background=None,
          parallel=True, cortex='classic', title=None, smoothing_steps=None,
          mask=True, subjects_dir=None, name=None, pos=None):
    """Create a :class:`Brain` object with a data layer

    Parameters
    ----------
    src : NDVar ([case,] source, [time]) | SourceSpace | str
        Data to plot; can be specified as :class:`NDVar` with source-space data,
        :class:`SourceSpace` dimension, or as subject name (:class:`str`).
        If ``src`` contains a ``Case`` dimension, the average across cases is
        taken. If it contains integer data, it is plotted as annotation,
        otherwise as data layer.
    cmap : str | array
        Colormap (name of a matplotlib colormap) or LUT array. If ``src`` is an
        integer NDVar, ``cmap`` can be a color dictionary mapping label IDs to
        colors.
    vmin, vmax : scalar
        Endpoints for the colormap. Need to be set explicitly if ``cmap`` is
        a LUT array.
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : str | sequence of str
        One or several views to show in the figure. The options are:
        ``'lateral', 'medial', 'ventral', 'dorsal', 'rostral', 'parietal',
        'frontal', 'caudal'``.
    hemi : 'lh' | 'rh' | 'both' | 'split'
        Which hemispheres to plot (default based on data).
    colorbar : bool
        Add a colorbar to the figure (use ``.plot_colorbar()`` to plot a
        colorbar separately).
    time_label : str
        Label to show time point. Use ``'ms'`` or ``'s'`` to display time in
        milliseconds or in seconds, or supply a custom format string to format
        time values (in seconds; default is ``'ms'``).
    w, h, axw, axh : scalar
        Layout parameters (figure width/height, subplot width/height).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    title : str
        title for the window (default is based on the subject name and
        ``src``).
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``src``. Can be matplotlib color, including
        alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white). If
        smoothing is enabled through ``smoothing_steps``, the mask is added as
        data layer, otherwise it is added as label. To add a mask independently,
        use the :meth:`Brain.add_mask` method.
    subjects_dir : None | str
        Override the subjects_dir associated with the source space dimension.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.
    pos : tuple of int
        Position of the new window on the screen.

    Returns
    -------
    brain : Brain
        Brain instance containing the plot.
    """
    if not use_inline_backend():
        from .._wxgui import get_app
        get_app(jumpstart=True)
    from ._brain_object import Brain, get_source_dim

    if isinstance(src, SourceSpace):
        if cmap is not None:
            raise TypeError(f"{cmap=} when plotting SourceSpace without data")
        elif vmin is not None:
            raise TypeError(f"{vmin=} when plotting SourceSpace without data")
        elif vmax is not None:
            raise TypeError(f"{vmax=} when plotting SourceSpace without data")
        ndvar = None
        source = src
        subject = source.subject
    elif isinstance(src, str):
        subject = src
        subjects_dir = mne.utils.get_subjects_dir(subjects_dir, True)
        ndvar = None
        source = None
        mask = False
        if hemi is None:
            hemi = 'split'
    else:
        ndvar = brain_data(src)
        if ndvar.has_case:
            ndvar = ndvar.summary()
        source = get_source_dim(ndvar)
        subject = source.subject
        # check that ndvar has the right dimensions
        if ndvar.ndim == 2 and not ndvar.has_dim('time') or ndvar.ndim > 2:
            raise ValueError(f"{ndvar=}: should have dimesions source and optionally time")
        if title is None and name is None and ndvar.name is not None:
            title = f"{source.subject} - {ndvar.name}"

    if hemi is None:
        if source.lh_n and source.rh_n:
            hemi = 'split'
        elif source.lh_n:
            hemi = 'lh'
        elif not source.rh_n:
            raise ValueError('No data')
        else:
            hemi = 'rh'
    elif source is None:
        pass
    elif (hemi == 'lh' and source.rh_n) or (hemi == 'rh' and source.lh_n):
        if ndvar is None:
            source = source[source._array_index(hemi)]
        else:
            ndvar = ndvar.sub(**{source.name: hemi})
            source = ndvar.get_dim(source.name)

    if subjects_dir is None:
        subjects_dir = source.subjects_dir

    brain = Brain(subject, hemi, surf, title, cortex, views=views, w=w, h=h, axw=axw, axh=axh, foreground=foreground, background=background, subjects_dir=subjects_dir, name=name, pos=pos, source_space=source)

    if ndvar is not None:
        if ndvar.x.dtype.kind == 'b':
            brain.add_ndvar_label(ndvar, cmap, False)
        elif ndvar.x.dtype.kind in 'ui':
            brain.add_ndvar_annotation(ndvar, cmap, False)
        else:
            brain.add_ndvar(ndvar, cmap, vmin, vmax, smoothing_steps, colorbar, time_label)

    if mask is not False:
        brain.add_mask(source, mask, smoothing_steps, None, subjects_dir)

    if parallel:
        brain.set_parallel_view(scale=True)
    return brain


@deprecated("0.25", brain)
def surfer_brain(*args, **kwargs):
    pass


def _voxel_brain(data, lut, vmin, vmax):
    """Plot spheres for volume source space

    Parameters
    ----------
    data : NDVar
        Data to plot.
    lut : array
        Color LUT.
    vmin, vmax : scalar
        Data range.
    """
    if data.dimnames != ('source',):
        raise ValueError("Can only plot 1 dimensional source space NDVars")

    from mayavi import mlab

    x, y, z = data.source.coordinates.T

    figure = mlab.figure()
    mlab.points3d(x, y, z, scale_factor=0.002, opacity=0.5)
    pts = mlab.points3d(x, y, z, data.x, vmin=vmin, vmax=vmax)
    pts.module_manager.scalar_lut_manager.lut.table = lut
    return figure


################################################################################
# Bin-Tables
############
# Top-level functions for fmtxt image tables and classes for Eelfigures.
# - _x_bin_table_ims() wrap 'x' brain plot function
# - _bin_table_ims() creates ims given a brain plot function

class ImageTable(ColorBarMixin, EelFigure):
    # Initialize in two steps
    #
    #  1) Initialize class to generate layout
    #  2) Use ._res_h and ._res_w to generate images
    #  3) Finalize by calling ._add_ims()
    #

    def __init__(
            self,
            n_rows: int,
            n_columns: int,
            row_labels: list = False,
            column_labels: list = False,
            title: str = None,
            margins: dict = None,
            **kwargs,
    ):
        default_margins = {}
        if row_labels:
            default_margins['left'] = 0.3
        layout = ImLayout(n_rows * n_columns, 4 / 3, 2, margins, default_margins, title, column_labels, rows=n_rows, columns=n_columns, autoscale=True, **kwargs)
        EelFigure.__init__(self, None, layout)

        self._n_rows = n_rows
        self._n_columns = n_columns
        fig_dpi = matplotlib.rcParams['savefig.dpi']
        if isinstance(fig_dpi, str):
            dpi = layout.dpi
        else:
            dpi = max(layout.dpi, fig_dpi)
        self._res_w = int(round(layout.axw * dpi))
        self._res_h = int(round(layout.axh * dpi))
        if row_labels:
            self.add_row_titles(row_labels, rotation='vertical')
        if column_labels:
            self.add_column_titles(column_labels)

    def _add_ims(
            self,
            ims: Sequence,  # shape = (rows, columns)
            cmap_params,
            cmap_data,
    ):
        for row, column in product(range(self._n_rows), range(self._n_columns)):
            ax = self.axes[row * self._n_columns + column]
            ax.imshow(ims[row][column])

        ColorBarMixin.__init__(self, lambda: cmap_params, cmap_data)
        self._show()

    def _fill_toolbar(self, tb):
        ColorBarMixin._fill_toolbar(self, tb)

    def add_column_titles(self, titles, x=None, y=0.25, va='center', ha='center', **kwargs):
        """Add a title for each column of images

        Parameters
        ----------
        titles : sequence of str
            Titles, from top to bottom.
        x : scalar
            Horizontal distance from left of the axes (default is to center
            over the axes).
        y : scalar
            Vertical distance from the top of the figure.
        ...
            Matplotlib text parameters.
        """
        if len(titles) > self._n_columns:
            raise ValueError(f"titles={titles}: {len(titles)} titles for {self._n_columns} columns")
        # distribute titles over columns
        n = self._n_columns / len(titles)  # columns per title
        column_width = self._layout.axw * n + self._layout.margins['wspace'] * (n - 1)
        # center
        if x is None:
            x = column_width / 2
        first = self._layout.margins['left'] + x
        y_ = 1 - y / self._layout.h
        for i, label in enumerate(titles):
            x_ = (first + i * column_width) / self._layout.w
            self.figure.text(x_, y_, label, va=va, ha=ha, **kwargs)
        self.draw()

    def add_row_titles(self, titles, x=0.1, y=None, va='center', ha='center', **kwargs):
        """Add a title for each row of images

        Parameters
        ----------
        titles : sequence of str
            Titles, from top to bottom.
        x : scalar
            Horizontal distance from left of the figure.
        y : scalar
            Vertical distance from the bottom of the axes (default is the
            center of the axes).
        ...
            Matplotlib text parameters.
        """
        if len(titles) > self._n_rows:
            raise ValueError(f"titles={titles}: {len(titles)} titles for {self._n_rows} rows")
        # distribute titles over columns
        n = self._n_rows / len(titles)  # columns per title
        row_height = self._layout.axh * n + self._layout.margins['hspace'] * (n - 1)
        # center
        if y is None:
            y = row_height / 2
        y_top = self._layout.margins['top'] + y
        x_ = x / self._layout.w
        for i, label in enumerate(titles):
            y_ = 1 - (y_top + i * row_height) / self._layout.h
            self.figure.text(x_, y_, label, va=va, ha=ha, **kwargs)
        self.draw()


class _BinTable(EelFigure, ColorBarMixin):
    """Super-class"""
    def __init__(self, ndvar, tstart, tstop, tstep, im_func, surf, views, hemi,
                 summary, title, foreground=None, background=None,
                 parallel=True, smoothing_steps=None, mask=True, margins=None,
                 **kwargs):
        if isinstance(views, str):
            views = (views,)
        data = ndvar.bin(tstep, tstart, tstop, summary)
        n_columns = len(data.time)
        n_hemis = (data.source.lh_n > 0) + (data.source.rh_n > 0)
        n_rows = len(views) * n_hemis

        layout = ImLayout(n_rows * n_columns, 4 / 3, 2, margins, {'bottom': 0.5}, title, rows=n_rows, columns=n_columns, **kwargs)
        EelFigure.__init__(self, None, layout)

        res_w = int(layout.axw * layout.dpi)
        res_h = int(layout.axh * layout.dpi)
        ims, header, cmap_params = im_func(data, surf, views, hemi, axw=res_w,
                                           axh=res_h, foreground=foreground,
                                           background=background,
                                           parallel=parallel,
                                           smoothing_steps=smoothing_steps,
                                           mask=mask)
        for row in range(n_rows):
            for column in range(n_columns):
                ax = self.axes[row * n_columns + column]
                ax.imshow(ims[row][column])

        # time labels
        y = 0.25 / layout.h
        for i, label in enumerate(header):
            x = (0.5 + i) / layout.columns
            self.figure.text(x, y, label, va='center', ha='center')

        ColorBarMixin.__init__(self, lambda: cmap_params, data)
        self._show()

    def _fill_toolbar(self, tb):
        ColorBarMixin._fill_toolbar(self, tb)


class BinTable(_BinTable):
    """DSPM plot bin-table"""
    def __init__(self, ndvar, tstart=None, tstop=None, tstep=0.1,
                 fmin=13, fmax=22, fmid=None,
                 surf='smoothwm', views=('lat', 'med'), hemi=None,
                 summary='sum', title=None, *args, **kwargs):
        im_func = partial(_dspm_bin_table_ims, fmin, fmax, fmid)
        _BinTable.__init__(self, ndvar, tstart, tstop, tstep, im_func, surf,
                           views, hemi, summary, title, *args, **kwargs)


class ClusterBinTable(_BinTable):
    """Data plotted on brain for different time bins and views

    Parameters
    ----------
    ndvar : NDVar (time x source)
        Data to be plotted.
    tstart : None | scalar
        Time point of the start of the first bin (inclusive; None to use the
        first time point in ndvar).
    tstop : None | scalar
        End of the last bin (exclusive; None to end with the last time point
        in ndvar).
    tstep : scalar
        Size of each bin (in seconds).
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : list of str
        Views to display (for each hemisphere, lh first). Options are:
        'rostral', 'parietal', 'frontal', 'ventral', 'lateral', 'caudal',
        'medial', 'dorsal'.
    hemi : 'lh' | 'rh' | 'both'
        Which hemispheres to plot (default based on data).
    summary : str
        How to summarize data in each time bin. Can be the name of a numpy
        function that takes an axis parameter (e.g., 'sum', 'mean', 'max') or
        'extrema' which selects the value with the maximum absolute value.
        Default is sum.
    vmax : scalar != 0
        Maximum value in the colormap. Default is the maximum value in the
        cluster.
    title : str
        Figure title.
    """
    def __init__(self, ndvar, tstart=None, tstop=None, tstep=0.1,
                 surf='smoothwm', views=('lat', 'med'), hemi=None,
                 summary='sum', vmax=None, title=None, *args, **kwargs):
        im_func = partial(_cluster_bin_table_ims, vmax)
        _BinTable.__init__(self, ndvar, tstart, tstop, tstep, im_func, surf,
                           views, hemi, summary, title, *args, **kwargs)


def dspm_bin_table(ndvar, fmin=2, fmax=8, fmid=None,
                   tstart=None, tstop=None, tstep=0.1, surf='smoothwm',
                   views=('lat', 'med'), hemi=None, summary='extrema',
                   axw=300, axh=250, *args, **kwargs):
    """Create a table with images for time bins

    Parameters
    ----------
    ndvar : NDVar (time x source)
        Data to be plotted.
    fmin, fmax : scalar >= 0
        Start- and end-point for the color gradient for positive values. The
        gradient for negative values goes from -fmin to -fmax. Values between
        -fmin and fmin are transparent.
    fmid : None | scalar
        Midpoint for the color gradient. If fmid is None (default) it is set
        half way between fmin and fmax.
    tstart : None | scalar
        Time point of the start of the first bin (inclusive; None to use the
        first time point in ndvar).
    tstop : None | scalar
        End of the last bin (exclusive; None to end with the last time point
        in ndvar).
    tstep : scalar
        Size of each bin (in seconds).
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : list of str
        Views to display (for each hemisphere, lh first). Options are:
        'rostral', 'parietal', 'frontal', 'ventral', 'lateral', 'caudal',
        'medial', 'dorsal'.
    hemi : 'lh' | 'rh' | 'both'
        Which hemispheres to plot (default based on data).
    summary : str
        How to summarize data in each time bin. Can be the name of a numpy
        function that takes an axis parameter (e.g., 'sum', 'mean', 'max') or
        'extrema' which selects the value with the maximum absolute value.
        Default is extrema.
    axw, axh : scalar
        Subplot width/height (default axw=300, axh=250).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``ndvar``. Can be matplotlib color,
        including alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white).
    subjects_dir : None | str
        Override the subjects_dir associated with the source space dimension.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.


    Returns
    -------
    images : Image
        FMTXT Image object that can be saved as SVG or integrated into an
        FMTXT document.


    Notes
    -----
    Plotting is based on :func:`plot.brain.dspm`.

    The resulting image can be saved with either of the following (currently
    the only supported formats)::

    >>> image.save_html("name.html")
    >>> image.save_image("name.svg")


    See Also
    --------
    plot.brain.bin_table: plotting clusters as bin-table
    """
    data = ndvar.bin(tstep, tstart, tstop, summary)
    ims, header, _ = _dspm_bin_table_ims(fmin, fmax, fmid, data, surf, views,
                                         hemi, axw, axh, *args, **kwargs)
    return im_table(ims, header)


def bin_table(ndvar, tstart=None, tstop=None, tstep=0.1, surf='smoothwm',
              views=('lat', 'med'), hemi=None, summary='sum', vmax=None,
              axw=300, axh=250, *args, **kwargs):
    """Create a table with images for time bins

    Parameters
    ----------
    ndvar : NDVar (time x source)
        Data to be plotted.
    tstart : None | scalar
        Time point of the start of the first bin (inclusive; None to use the
        first time point in ndvar).
    tstop : None | scalar
        End of the last bin (exclusive; None to end with the last time point
        in ndvar).
    tstep : scalar
        Size of each bin (in seconds).
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : list of str
        Views to display (for each hemisphere, lh first). Options are:
        'rostral', 'parietal', 'frontal', 'ventral', 'lateral', 'caudal',
        'medial', 'dorsal'.
    hemi : 'lh' | 'rh' | 'both'
        Which hemispheres to plot (default based on data).
    summary : str
        How to summarize data in each time bin. Can be the name of a numpy
        function that takes an axis parameter (e.g., 'sum', 'mean', 'max') or
        'extrema' which selects the value with the maximum absolute value.
        Default is sum.
    vmax : scalar != 0
        Maximum value in the colormap. Default is the maximum value in the
        cluster.
    out : 'image' | 'figure'
        Format in which to return the plot. ``'image'`` (default) returns an
        Image object that
    axw, axh : scalar
        Subplot width/height (default axw=300, axh=250).
    foreground : mayavi color
        Figure foreground color (i.e., the text color).
    background : mayavi color
        Figure background color.
    parallel : bool
        Set views to parallel projection (default ``True``).
    cortex : str | tuple | dict
        Mark gyri and sulci on the cortex. Presets: ``'classic'`` (default),
        ``'high_contrast'``, ``'low_contrast'``, ``'bone'``. Can also be a
        single color (e.g. ``'red'``, ``(0.1, 0.4, 1.)``) or a tuple of two
        colors for gyri and sulci (e.g. ``['red', 'blue']`` or ``[(1, 0, 0),
        (0, 0, 1)]``). For all options see the PySurfer documentation.
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``ndvar``. Can be matplotlib color,
        including alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white).
    subjects_dir : None | str
        Override the subjects_dir associated with the source space dimension.
    name : str
        Equivalent to ``title``, for consistency with other plotting functions.


    Returns
    -------
    images : Image
        FMTXT Image object that can be saved as SVG or integrated into an
        FMTXT document.


    Notes
    -----
    Plotting is based on :func:`plot.brain.cluster`.

    The resulting image can be saved with either of the following (currently
    the only supported formats)::

    >>> image.save_html("name.html")
    >>> image.save_image("name.svg")


    See Also
    --------
    plot.brain.dspm_bin_table: plotting SPMs as bin-table
    """
    data = ndvar.bin(tstep, tstart, tstop, summary)
    ims, header, _ = _cluster_bin_table_ims(vmax, data, surf, views, hemi,
                                            axw, axh, *args, **kwargs)
    return im_table(ims, header)


def _dspm_bin_table_ims(fmin, fmax, fmid, data, surf, views, hemi, axw, axh,
                        *args, **kwargs):
    if fmax is None:
        fmax = max(data.max(), -data.min())

    def brain_(hemi_):
        return dspm(data, fmin, fmax, fmid, surf, views[0], hemi_, False, None,
                    axw, axh, *args, **kwargs)

    return _bin_table_ims(data, hemi, views, brain_)


def _cluster_bin_table_ims(vmax, data, surf, views, hemi, axw, axh, *args,
                           **kwargs):
    if vmax is None:
        vmax = max(-data.min(), data.max())
        if vmax == 0:
            raise NotImplementedError("The data of ndvar is all zeros")
    elif vmax == 0:
        raise ValueError("vmax can't be 0")
    elif vmax < 0:
        vmax = -vmax

    def brain_(hemi_):
        return cluster(data, vmax, surf, views[0], hemi_, False, None, axw,
                       axh, None, None, *args, **kwargs)

    return _bin_table_ims(data, hemi, views, brain_)


def _bin_table_ims(data, hemi, views, brain_func):
    if isinstance(views, str):
        views = (views,)
    ims = []
    if hemi is None:
        hemis = []
        if data.source.lh_n:
            hemis.append('lh')
        if data.source.rh_n:
            hemis.append('rh')
    elif hemi == 'both':
        hemis = ['lh', 'rh']
    elif hemi == 'lh' or hemi == 'rh':
        hemis = [hemi]
    else:
        raise ValueError("hemi=%s" % repr(hemi))

    cmap_params = None
    for hemi in hemis:
        brain = brain_func(hemi)

        hemi_lines = [[] for _ in views]
        for i in range(len(data.time)):
            brain.set_data_time_index(i)
            for line, view in zip(hemi_lines, views):
                brain.show_view(view)
                im = brain.screenshot_single('rgba', True)
                line.append(im)
        ims += hemi_lines
        if cmap_params is None:
            cmap_params = brain._get_cmap_params()
        brain.close()

    header = ['%i - %i ms' % (ms(t0), ms(t1)) for t0, t1 in data.info['bins']]
    return ims, header, cmap_params


class SPLayer(Enum):
    UNDEFINED = 0
    ITEM = 1  # multiple separate NDVars
    SEQUENCE = 2  # One NDVar with multiple images (e.g. time sequence)
    OVERLAY = 3  # static overlay applying to all items/the whole sequence


class SPPlotType(Enum):
    DATA = 1
    LABEL = 2
    MNE_LABEL = 3
    FUNC = 4


@dataclass
class SequencePlotterLayer:
    kind: SPLayer
    data: Union[NDVar, mne.Label, mne.BiHemiLabel]
    args: tuple
    kwargs: dict
    label: Sequence[str] = None
    index: int = None  # when adding multiple items, index of the item
    plot_type: SPPlotType = SPPlotType.DATA

    def __post_init__(self):
        if self.label is not None:
            if self.kind == SPLayer.SEQUENCE:
                if not isinstance(self.label, (tuple, list)):
                    raise TypeError(f"label={self.label!r}; list or tuple of str expected")
            else:
                if not isinstance(self.label, str):
                    raise TypeError(f"label={self.label!r}; str expected")
        if self.kind == SPLayer.ITEM:
            assert isinstance(self.index, int)
        else:
            assert self.index is None

    def plot(self, brain):
        if self.plot_type == SPPlotType.MNE_LABEL:
            if isinstance(self.data, mne.BiHemiLabel):
                label = getattr(self.data, brain.hemi)
            elif self.data.hemi == brain.hemi:
                label = self.data
            else:
                return
            brain.add_label(label, **self.kwargs)
        elif self.plot_type == SPPlotType.FUNC:
            func = self.args[0]
            func(brain)
        elif len(self.data.source.vertices[brain.hemi == 'rh']) == 0:
            return
        elif self.plot_type == SPPlotType.DATA:
            brain.add_ndvar(self.data, *self.args, time_label='', **self.kwargs)
        elif self.plot_type == SPPlotType.LABEL:
            brain.add_ndvar_label(self.data, *self.args, **self.kwargs)
        else:
            raise RuntimeError(f"{self.plot_type=}")


class SequencePlotter:
    """Grid of anatomical images in one figure

    Notes
    -----
    For plots with multiple layers, layers are added to the brain in the order
    they were added to the ``SequencePlotter``.

    Examples
    --------
    Plot an evoked response or TRF (i.e., a time by source :class:`NDVar`)
    in 50 ms bins between 0 and 300 ms. Use the :meth:`~NDVar.extrema` method to
    plot peak activity in each time bin. This will plot the most extreme
    value including its sign, and thus show positive and negative effects::

        # Bin the data
        ndvar_binned = ndvar.bin(step=0.050, start=0, stop=0.3, func='extrema')
        # Initialize the SequencePlotter and set parameters for visualizing the brain
        sp = plot.brain.SequencePlotter()
        sp.set_brain_args(surf='smoothwm')
        # Add the data to be plotted
        sp.add_ndvar(ndvar_binned)
        # Generate the figure with multiple brain plots
        p = sp.plot_table(view='lateral')
        # Save as you would normally
        p.save('Figure.pdf')

    Plot specific time points in an evoked response or TRF (``ndvar``). To do
    this, simply add mutiple data objects. Make sure to pass a colormap or
    ``vmax`` parameter to make the color-scale consistent between plots::

        cmap = plot.soft_threshold_colormap('xpolar-a', 0.0001, 0.010)
        sp = plot.brain.SequencePlotter()
        sp.set_brain_args(surf='smoothwm')
        for t in [0.050, 0.100, 0.200]:
            sp.add_ndvar(ndvar.sub(time=t), cmap=cmap, label=f'{int(t*1000)} ms')
        p = sp.plot_table(view='lateral', orientation='vertical')

    Visualize a test result, by separately adding condition means and a
    difference map that is masked by significance::

        res = testnd.TTestRelated('srcm', 'condition', 'a', 'b', match='subject', data=data)
        vmax = 3  # explicitly set vmax to make sure that the color-maps agree
        sp = plot.brain.SequencePlotter()
        sp.set_brain_args(surf='inflated')
        sp.add_ndvar(res.c1_mean, vmax=vmax, label='a')
        sp.add_ndvar(res.c0_mean, vmax=vmax, label='b')
        sp.add_ndvar(res.masked_difference(), vmax=vmax, label='a - b')
        p = sp.plot_table(view='lateral', orientation='vertical')

    Plot a source by time test result, showing how a difference map evolves
    over time::

        res = testnd.TTestRelated('srcm', 'condition', 'a', 'b', match='subject', data=data)
        difference = res.masked_difference()
        binned = difference.bin(step=0.200, start=0.200, stop=1.00, func='extrema')
        sp = plot.brain.SequencePlotter()
        sp.add_ndvar(binned)
        p = sp.plot_table(view='lateral', title='a = b')

    """
    max_n_bins = 25

    def __init__(
            self,
            subject: str = None,
            subjects_dir: PathArg = None,
    ):
        self._data = []
        self._source = None  # for mask
        self._subject = subject
        self._subjects_dir = subjects_dir
        self._frame_dim = None
        self._frame_labels = None
        self._bin_kind: SPLayer = SPLayer.UNDEFINED
        self._frame_order = None
        self._brain_args = {}
        self._parallel_view = {}

    def _assert_has_frame_dim(self):
        if self._frame_dim is None:
            raise RuntimeError("Need to set at least one NDVar first")
        elif self._frame_dim is False:
            raise RuntimeError("Only applies to SequencePlotters with 2d NDVars")

    def set_brain(
            self,
            source: SourceSpace = None,
            subject: str = None,
            subjects_dir: PathArg = None,
    ):
        """Set the brain model on which to plot

        This is usually handled automatically, and only needs to be invoked
        directly for plots not otherwise involving an :class:`NDVar` with a
        :class:`SourceSpace` dimension.

        Parameters
        ----------
        source
            Brain model on which to plot. The coverage if the source space is
            also used for the mask (the first source space that is added takes
            precedence).
        subject
            Specify subject directly.
        subjects_dir
            Specify the subjects directory directly.
        """
        if isinstance(source, SourceSpace):
            subject = source.subject
            subjects_dir = source.subjects_dir
        elif source is not None:
            raise TypeError(f"{source!r}")

        if self._subject is None:
            self._subject = subject
        elif subject and subject != self._subject:
            raise ValueError(f"Different subject ({subject}) than previously added ({self._subject})")
        if self._subjects_dir is None:
            self._subjects_dir = subjects_dir
        elif subjects_dir and subjects_dir != self._subjects_dir:
            raise ValueError(f"different subjects_dir ({subjects_dir}) than previously added ({self._subjects_dir})")

        if self._source is None and isinstance(source, SourceSpace):
            self._source = source

    def _n_items(self):
        if self._bin_kind == SPLayer.SEQUENCE:
            raise RuntimeError("Trying to retrieve number of items for dimension-based plot")
        indices = [l.index for l in self._data if l.index is not None]
        if indices:
            return max(indices) + 1
        self._bin_kind = SPLayer.ITEM
        return 0

    def set_brain_args(self, surf='inflated', foreground=None, background=None, parallel=True, cortex='classic', mask=True):
        "Set parameters for anatomical plot (see :func:`.brain`)"
        self._brain_args = {'surf': surf, 'foreground': foreground, 'background': background, 'parallel': parallel, 'cortex': cortex, 'mask': mask}

    def set_parallel_view(self, forward=None, up=None, scale=None):
        "Set view for all plots (see :meth:`~.brain_object.Brain.set_parallel_view`"
        if forward is not None:
            self._parallel_view['forward'] = forward
        if up is not None:
            self._parallel_view['up'] = up
        if scale is not None:
            self._parallel_view['scale'] = scale

    def _index(self, index: int) -> int:
        # index arg to valid index
        if index is None:
            return self._n_items()
        elif index < 0:
            return self._n_items() + index
        else:
            return index

    def add_function(self, func, label=None, overlay=False):
        "Custom modification of the brain object (calls ``func(brain)``)"
        args = (func,)
        if overlay:
            index = None
            kind = SPLayer.OVERLAY
        else:
            index = self._n_items()
            kind = SPLayer.ITEM
        layer = SequencePlotterLayer(kind, None, args, {}, label, index, SPPlotType.FUNC)
        self._data.append(layer)

    def add_label(
            self,
            mne_label: Union[mne.Label, mne.BiHemiLabel],
            *,
            overlay: bool = False,
            index: int = None,
            label: str = None,
            **kwargs,
    ):
        """Add a label to the plot.

        Parameters
        ----------
        overlay
            Plot the label as overlay over all plots in the sequence.
        index
            When adding separate frames (rows/columns), explicitly specify the
            frame index (the default is to add a frame at the end; use -1 to
            add an overlay to the most recent frame).
        label
            Label when adding multiple separate NDVars. Labels for bins when
            adding ``ndvar`` with multiple bins.
        ...
            See :meth:`~._brain_object.Brain.add_label`
        """
        self.set_brain(subject=mne_label.subject)
        if overlay:
            kind = SPLayer.OVERLAY
        else:
            index = self._index(index)
            kind = SPLayer.ITEM
        layer = SequencePlotterLayer(kind, mne_label, (), kwargs, label, index, SPPlotType.MNE_LABEL)
        self._data.append(layer)

    def add_ndvar(
            self,
            ndvar: NDVar,
            *args,
            static: bool = None,
            index: int = None,
            label: Union[str, Sequence[str]] = None,
            **kwargs,
    ):
        """Add a data layer to the brain plot

        Multiple data layers can be added sequentially, but each additional
        layer needs to have a time dimension that is compatible with previous
        layers (or no time dimension).

        Parameters
        ----------
        ndvar
            Data to add. ``Source`` dimension only for a static layer,
            additional ``time`` or ``case`` dimension for dynamic layers.
        ...
            :meth:`~._brain_object.Brain.add_ndvar` parameters.
        static
            When supplying a one dimensional ``ndvar``, whether to consider it
            as a static layer (an overlay appearing on all plots) or as one of
            several plots (default).
        index
            When adding separate frames (rows/columns), explicitly specify the
            frame index (the default is to add a frame at the end; use -1 to
            add an overlay to the most recent frame).
        label
            Label when adding multiple separate NDVars. Labels for bins when
            adding ``ndvar`` with multiple bins.
        """
        self.set_brain(ndvar.get_dim('source'))

        # find row/column dimension
        if ndvar.ndim == 1:
            frame_dim = None
        elif ndvar.ndim == 2:
            dim_name = ndvar.get_dimnames((None, 'source'))[0]
            frame_dim = ndvar.get_dim(dim_name)
        elif ndvar.has_case:
            raise ValueError(f"{ndvar=}: must be one- or two dimensional. If you meant to plot the average of cases, use ndvar.mean('case')")
        else:
            raise ValueError(f"{ndvar=}: must be one- or two dimensional")

        if frame_dim is None:
            # apply defaults
            if static is None:
                static = True if self._frame_dim else False  # None: unset or use multiple NDVars
            # extract
            if static:
                kind = SPLayer.OVERLAY
            elif self._frame_dim:
                raise ValueError(f"{static=} when adding to plot with dimension {self._frame_dim}")
            else:
                kind = SPLayer.ITEM
                self._bin_kind = SPLayer.ITEM
                index = self._index(index)
            layer = SequencePlotterLayer(kind, ndvar, args, kwargs, label, index)
        elif self._bin_kind == SPLayer.ITEM:
            raise ValueError("Can't add 2d NDVar to SequencePlotter with multiple NDVars")
        elif index is not None:
            raise TypeError(f"{index=}")
        elif static:
            raise ValueError(f"{static=} for multi-dimensional NDVar")
        else:
            self._bin_kind = SPLayer.SEQUENCE
            if self._frame_dim is None:
                if len(frame_dim) > self.max_n_bins:
                    raise ValueError(f"Trying to plot {frame_dim} with {len(frame_dim)} bins. If this is intentional, set SequencePlotter.max_n_bins to a larger value.")
                self._frame_dim = frame_dim
            elif frame_dim != self._frame_dim:
                raise ValueError(f"New axis {frame_dim} is incompatible with previously set axis {self._frame_dim}")
            if self._frame_labels is None and 'bins' in ndvar.info:
                bins = ndvar.info['bins']
                if isinstance(bins[0], str):
                    self._frame_labels = bins
                elif isinstance(frame_dim, UTS):
                    fmt, _, unit = frame_dim._axis_format(True, '__unit__')
                    self._frame_labels = [f'{t0 * 1000:.0f} - {t1 * 1000:.0f}' for t0, t1 in bins]
                    self._frame_labels[0] = f'{self._frame_labels[0]} {unit}'

            layer = SequencePlotterLayer(SPLayer.SEQUENCE, ndvar, args, kwargs, label)

        self._data.append(layer)

    def add_ndvar_label(
            self,
            ndvar: NDVar,
            color: Any = (1, 0, 0),
            borders: int = False,
            name: str = None,
            alpha: float = None,
            lighting: bool = False,
            overlay: bool = False,
            index: int = None,
            label: str = None,
    ):
        """Add a boolean label to the brain plot

        Parameters
        ----------
        ...
            See :meth:`~._brain_object.Brain.add_ndvar_label`
        overlay
            Plot the label as overlay over all plots in the sequence.
        index
            When adding separate frames (rows/columns), explicitly specify the
            frame index (the default is to add a frame at the end; use -1 to
            add an overlay to the most recent frame).
        label
            Label when adding multiple separate NDVars. Labels for bins when
            adding ``ndvar`` with multiple bins.
        """
        self.set_brain(ndvar.get_dim('source'))
        args = (color, borders, name, alpha, lighting)
        if overlay:
            kind = SPLayer.OVERLAY
        else:
            index = self._index(index)
            kind = SPLayer.ITEM
        layer = SequencePlotterLayer(kind, ndvar, args, {}, label, index, SPPlotType.LABEL)
        self._data.append(layer)

    def add_pmap(self, res: Union[NDTest, NDVar], label=None, **kwargs):
        "See :meth:`~._brain_object.Brain.add_ndvar_p_map`"
        # interpret NDTest subtypes
        if isinstance(res, MultiEffectNDTest):
            iterator = zip(res.p, res._statistic_map, res.effects)
        elif isinstance(res, NDVar):
            iterator = ([res, None, res.name],)
        else:
            if isinstance(res, NDDifferenceTest):
                desc = res.difference.name
            else:
                desc = res._name()
            iterator = [(res.p, res._statistic_map, desc)]
        # add p-maps
        for p_map, stat_map, desc in iterator:
            ndvar, cmap, vmax = p_lut(p_map, stat_map)
            layer_label = desc if label is None else label
            self.add_ndvar(ndvar, cmap, -vmax, vmax, label=layer_label, **kwargs)

    def set_frame_order(self, order):
        """Set the order in which frames are plotted

        Parameters
        ----------
        order : list
            Sequence of frame dimension indices.
        """
        self._assert_has_frame_dim()
        missing = [x for x in order if x not in self._frame_dim]
        if missing:
            raise ValueError(f"order={order!r}; the following elements are not part of the frame dimension: {', '.join(missing)}")
        self._frame_order = order

    def _get_frame_labels(self, labels):
        if labels in (None, False):
            return
        elif labels is True:
            if self._bin_kind == SPLayer.ITEM:
                labels = []
                for i in range(self._n_items()):
                    layers = [l for l in self._data if l.index == i]
                    items = [l.label for l in layers if l.label is not None]
                    if len(items) == 0:
                        label = ''
                    else:
                        label = items[0]
                    labels.append(label)
            else:
                if self._frame_labels is not None:
                    return self._frame_labels
                elif self._frame_dim is None:
                    labels = []
                elif self._frame_order is None:
                    labels = list(self._frame_dim)
                else:
                    index = self._frame_dim._array_index(self._frame_order)
                    labels = [self._frame_dim[i] for i in index]
            if not any(labels):
                return

        # float -> ms
        out = []
        for label in labels:
            if isinstance(label, str):
                out.append(label)
            elif isinstance(label, Number):
                out.append(f'{ms(label)} ms')
            elif len(label) == 2:
                out.append(f'{ms_window(*label)} ms')
            else:
                out.append(str(label))
        return out

    @property
    def _bins(self):
        if self._frame_order:
            return self._frame_dim._array_index(self._frame_order)
        elif self._frame_dim:
            return list(range(len(self._frame_dim)))
        elif self._bin_kind == SPLayer.ITEM:
            return list(range(self._n_items()))
        else:
            return [None]

    def plot_table(
            self,
            hemi: Literal['lh', 'rh', 'both'] = None,
            view: Union[str, Sequence[str]] = ('lateral', 'medial'),
            orientation: Literal['horizontal', 'vertical'] = None,
            labels: Union[bool, Sequence[str]] = True,
            mode: Literal['rgb', 'rgba'] = 'rgb',
            antialiased: bool = False,
            rows: Sequence[str] = ('view',),
            columns: Sequence[str] = ('bin', 'hemi'),
            hemi_magnet: float = 0,
            **kwargs,
    ) -> ImageTable:
        """Create a figure with the images

        Parameters
        ----------
        hemi
            Hemispheres to plot (default is all hemispheres with data).
        view
            Views to plot. A view can be specified as a string, or as a tuple
            including parallel-view parameters ``(view, forward, up, scale)``,
            e.g., ``('lateral', 0, 10, 70)``.
        orientation
            Direction of the time/case axis; overrides ``row`` and ``column``
            settings.
        labels
            Headers for columns/rows of images (default is inferred from the data).
        mode
            Image mode (default ```'rgb'``, set to ``'rgba'`` to include alpha
            channel).
        antialiased
            Apply antialiasing to the images (default ``False``).
        rows
            What to alternate along the rows.
        columns
            What to alternate along the columns.
        hemi_magnet
            Pull the two hemispheres of the same plot closer together (movement
            in inches).
        ...
            Also accepts :ref:`general-layout-parameters`.

        Returns
        -------
        fig
            Figure created by the plot.

        Notes
        -----
        Use the ``margins`` parameter to control image spacing. For example (all
        numbers in inches)::

            margins = {
                'left': 0.1,    # left mergin
                'right': 0.1,   # right margin
                'top': 0.5,     # margin at the top, include space for title
                'bottom': 0.1,  # bottom
                'hspace': 0.1,  # height of the space between images
                'wspace': 0.1,  # width of the space between images
            }
        """
        if self._subject is None:
            raise RuntimeError("subject not set; specify when initializing SequencePlotter")
        elif self._subjects_dir is None:
            raise RuntimeError("subjects_dir not set; specify when initializing SequencePlotter")
        elif not self._data:
            raise RuntimeError("No data")
        labels = self._get_frame_labels(labels)
        # determine hemisphere(s) to plot
        if hemi is None:
            hemis = []
            sss = [data.data.source for data in self._data if isinstance(data.data, NDVar)]
            if any(ss.lh_n for ss in sss):
                hemis.append('lh')
            if any(ss.rh_n for ss in sss):
                hemis.append('rh')
            if not hemis:
                hemis = ('lh', 'rh')
        elif hemi == 'both':
            hemis = ('lh', 'rh')
        elif hemi in ('lh', 'rh'):
            hemis = (hemi,)
        else:
            raise ValueError(f"{hemi=}")
        # views
        if isinstance(view, str):
            views = (view,)
        elif len(view) > 1 and not isinstance(view[1], str):
            views = (view,)
        else:
            views = view

        # determine im-table layout
        if orientation:
            if orientation == 'vertical':
                rows, columns = ('bin',), ('view', 'hemi')
            elif orientation == 'horizontal':
                rows, columns = ('view',), ('bin', 'hemi')
            else:
                raise ValueError(f"{orientation=}")
        axis_sequences = {'bin': self._bins, 'hemi': hemis, 'view': views}
        ns = {key: len(values) for key, values in axis_sequences.items()}
        specified = set(columns).union(rows)
        if specified != set(ns) or len(columns) + len(rows) != 3:
            raise ValueError(f"{rows=}, {columns=}")
        n_columns = reduce(operator.mul, (ns[key] for key in columns))
        n_rows = reduce(operator.mul, (ns[key] for key in rows))
        # labels
        bin_in_column = 'bin' in columns
        bin_axis = columns if bin_in_column else rows
        if labels and (bin_pos := bin_axis.index('bin')):
            n_tile = reduce(operator.mul, (ns[key] for key in bin_axis[:bin_pos]))
            labels = labels * n_tile
        if bin_in_column:
            column_labels, row_labels = labels, None
        else:
            column_labels, row_labels = None, labels

        figure = ImageTable(n_rows, n_columns, row_labels, column_labels, **kwargs)
        hemi_shift = int(round(figure._layout.dpi * hemi_magnet))

        ims = self._generate_ims(figure._res_w, figure._res_h, views, mode, antialiased, hemi_shift, hemis)
        cmap_params, cmap_data = ims.pop('cmap', (None, None))

        # reshape im-order
        row_pattern = list(product(*[axis_sequences[key] for key in rows]))
        column_pattern = list(product(*[axis_sequences[key] for key in columns]))
        im_array = np.ndarray((n_rows, n_columns), object)
        for i_row, key_row in enumerate(row_pattern):
            key_dict = {key: value for key, value in zip(rows, key_row)}
            for i_col, key_col in enumerate(column_pattern):
                key_dict.update({key: value for key, value in zip(columns, key_col)})
                key = tuple([key_dict[key] for key in ['hemi', 'bin', 'view']])
                im_array[i_row, i_col] = ims[key]
        figure._add_ims(im_array, cmap_params, cmap_data)
        return figure

    def _generate_ims(
            self,
            w: int,
            h: int,
            views: Sequence[str],
            mode: str = 'rgba',
            antialiased: bool = True,
            hemi_shift: int = 0,
            hemis: Sequence[str] = ('lh', 'rh'),
    ) -> dict:
        bins = self._bins
        ims = {}
        source_space = self._source or self._subject
        for hemi in hemis:
            # plot brain
            b = brain(source_space, hemi=hemi, views='lateral', w=w, h=h, time_label='', subjects_dir=self._subjects_dir, **self._brain_args)

            if self._bin_kind == SPLayer.SEQUENCE:
                for l in self._data:
                    l.plot(b)
                    if 'cmap' not in ims and l.kind == SPLayer.SEQUENCE:
                        ims['cmap'] = b._get_cmap_params(), l.data

                # capture images
                for i in bins:
                    if i is not None:
                        b.set_data_time_index(i)
                        self._capture(b, views, mode, antialiased, ims, hemi, i, hemi_shift)
            elif self._bin_kind in (SPLayer.ITEM, SPLayer.UNDEFINED):  # UNDEFINED: has overlays only
                for i in bins:
                    for l in self._data:
                        if l.kind == SPLayer.OVERLAY or l.index == i:
                            l.plot(b)
                            if 'cmap' not in ims and l.plot_type == SPPlotType.DATA and l.kind == SPLayer.ITEM:
                                ims['cmap'] = b._get_cmap_params(), l.data
                    self._capture(b, views, mode, antialiased, ims, hemi, i, hemi_shift)
                    b.remove_data()
                    b.remove_labels()
            else:
                raise NotImplementedError(f"{self._bin_kind}")
            b.close()
        return ims

    def _capture(self, b, views, mode, antialiased, ims, hemi, i, shift):
        for view in views:
            if isinstance(view, str):
                b.show_view(view)
                if self._parallel_view:
                    b.set_parallel_view(**self._parallel_view)
            else:
                b.show_view(view[0])
                b.set_parallel_view(*view[1:])
            im = b.screenshot_single(mode, antialiased)
            if shift:
                if hemi == 'lh':
                    im = np.concatenate([im[:, -shift:], im[:, :-shift]], axis=1)
                else:
                    im = np.concatenate([im[:, shift:], im[:, :shift]], axis=1)
            ims[hemi, i, view] = im


def adjacency(source):
    """Plot source space adjacency

    Parameters
    ----------
    source : SourceSpace | NDVar
        Source space or NDVar containing the source space.

    Returns
    -------
    figure : mayavi Figure
        The figure.
    """
    from mayavi import mlab

    if isinstance(source, NDVar):
        source = source.get_dim('source')

    connections = source.adjacency()
    coords = source.coordinates
    x, y, z = coords.T

    figure = mlab.figure()
    src = mlab.pipeline.scalar_scatter(x, y, z, figure=figure)
    src.mlab_source.dataset.lines = connections
    lines = mlab.pipeline.stripper(src)
    mlab.pipeline.surface(lines, colormap='Accent', line_width=1, opacity=1.,
                          figure=figure)
    return figure


def copy(brain):
    "Copy the figure to the clip board"
    return brain.copy_screenshot()


def butterfly(y, cmap=None, vmin=None, vmax=None, surf='inflated',
              views='lateral', hemi=None,
              w=5, h=2.5, smoothing_steps=None, mask=True,
              xlim=None, name=None):
    """Shortcut for a Butterfly-plot with a time-linked brain plot

    Parameters
    ----------
    y : NDVar  ([case,] time, source)
        Data to plot; if ``y`` has a case dimension, the mean is plotted.
        ``y`` can also be a :mod:`~eelbrain.testnd` t-test result, in which
        case a masked parameter map is plotted (p ≤ 0.05).
    cmap : str | array
        Colormap (name of a matplotlib colormap).
    vmin : scalar
        Plot data range minimum.
    vmax : scalar
        Plot data range maximum.
    surf : 'inflated' | 'pial' | 'smoothwm' | 'sphere' | 'white'
        Freesurfer surface to use as brain geometry.
    views : str | sequence of str
        One or several views to show in the figure. The options are:
        ``'lateral', 'medial', 'ventral', 'dorsal', 'rostral', 'parietal',
        'frontal', 'caudal'``.
    hemi : 'lh' | 'rh'
        Plot only this hemisphere (the default is to plot all hemispheres with
        data in ``y``).
    w : scalar
        Butterfly plot width (inches).
    h : scalar
        Plot height (inches; applies to butterfly and brain plot).
    smoothing_steps : None | int
        Number of smoothing steps if data is spatially undersampled (pysurfer
        ``Brain.add_data()`` argument).
    mask : bool | matplotlib color
        Shade areas that are not in ``src``. Can be matplotlib color, including
        alpha (e.g., ``(1, 1, 1, 0.5)`` for semi-transparent white). If
        smoothing is enabled through ``smoothing_steps``, the mask is added as
        data layer, otherwise it is added as label. To add a mask independently,
        use the :meth:`Brain.add_mask` method.
    xlim : scalar | (scalar, scalar)
        Initial x-axis view limits as ``(left, right)`` tuple or as ``length``
        scalar (default is the full x-axis in the data).
    name : str
        The window title (default is y.name).

    Returns
    -------
    butterfly_plot : plot.Butterfly
        Butterfly plot.
    brain : Brain
        Brain plot.
    """
    hemis, bfly_data, brain_data = butterfly_data(y, hemi)

    if name is None:
        name = brain_data.name

    # butterfly-plot
    p = Butterfly(bfly_data, vmin=vmin, vmax=vmax, xlim=xlim, h=h, w=w, columns=1, name=name, color='black', ylabel=hemis, axtitle=False)

    # Brain plot
    brain_h = h * p._layout.dpi
    p_brain = brain(brain_data, cmap, vmin, vmax, surf, views, hemi, mask=mask, smoothing_steps=smoothing_steps, axh=brain_h, name=name)

    p._auto_position(p_brain)
    p.link_time_axis(p_brain)

    return p, p_brain



"""Terminal-based implementation of the Eelbrain ui functions."""
import os


def ask_saveas(title, messagefiletypes, defaultDir, defaultFile):
    msg = "%s (%s): " % (title, message)
    path = input(msg)
    path = os.path.expanduser(path)

    dirname = os.path.split(path)[0]
    if os.path.exists(path):
        if ask(title="File Exists. Overwrite?",
               message=repr(path)):
            return path
        else:
            return False
    elif os.path.exists(dirname):
        return path
    else:
        if ask(title="Directory does not exist. Create?",
               message=repr(dirname)):
            os.makedirs(dirname)
            return path
        else:
            return False


def ask_dir(title="Select Folder", message="Please Pick a Folder",
            must_exist=True):
    msg = "%s (%s): " % (title, message)
    path = input(msg)
    path = os.path.expanduser(path)
    if os.path.exists(path) and os.path.isdir(path):
        return path
    else:
        return False


def ask_file(title, message, filetypes, directory, mult):
    msg = "%s (%s): " % (title, message)
    path = input(msg)
    path = os.path.expanduser(path)
    if os.path.exists(path):
        return path
    else:
        return False


def ask(title="Overwrite File?",
        message="Duplicate filename. Do you want to overwrite?",
        cancel=False, default=True):
    print(title)
    print(message)
    c = ''
    while c not in ['y', 'n', 'c']:
        c = input("([y]es / [n]o / [c]ancel)")
    if c == 'y':
        return True
    elif c == 'n':
        return False
    else:
        return None


def ask_color(parent=None, default=None):
    c = input('Color = ')
    return eval(c)


def message(title, message=None, icon='i'):
    if icon:
        title = "%s: %s" % (icon, title)
    print(title)
    if message:
        print(message)


def copy_file(path):
    raise NotImplementedError


def copy_text(text):
    raise NotImplementedError



# https://packaging.python.org/en/latest/
# https://packaging.python.org/en/latest/guides/modernize-setup-py-project/
import os
import platform
from setuptools import setup, find_packages, Extension

import numpy as np

# Source distribution includes C code to allow installing without Cython
# https://cython.readthedocs.io/en/stable/src/userguide/source_files_and_compilation.html#distributing-cython-modules
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = False


IS_WINDOWS = os.name == 'nt'
IS_ARM = platform.machine().lower().startswith('arm')

# Cython extensions
base_args = {'define_macros': [("NPY_NO_DEPRECATED_API", "NPY_1_11_API_VERSION")]}
if IS_WINDOWS:
    open_mp_args = {
        **base_args,
        'extra_compile_args': '/openmp',
    }
elif IS_ARM:
    open_mp_args = {
        **base_args,
        'extra_compile_args': ['-Wno-unreachable-code', '-Xpreprocessor', '-fopenmp', '-O3'],
        'extra_link_args': ['-Xpreprocessor', '-fopenmp'],
    }
    base_args['extra_compile_args'] = ['-Wno-unreachable-code', '-O3']
else:
    open_mp_args = {
        **base_args,
        'extra_compile_args': ['-Wno-unreachable-code', '-fopenmp', '-O3', '-mavx'],
        'extra_link_args': ['-fopenmp'],
    }
    base_args['extra_compile_args'] = ['-Wno-unreachable-code', '-O3', '-mavx']
ext = '.pyx' if cythonize else '.c'
ext_cpp = '.pyx' if cythonize else '.cpp'
extensions = [
    Extension('eelbrain._data_opt', [f'eelbrain/_data_opt{ext}'], **base_args),
    Extension('eelbrain._trf._boosting_opt', [f'eelbrain/_trf/_boosting_opt{ext}'], **open_mp_args),
    Extension('eelbrain._ndvar._convolve', [f'eelbrain/_ndvar/_convolve{ext}'], **open_mp_args),
    Extension('eelbrain._ndvar._gammatone', [f'eelbrain/_ndvar/_gammatone{ext}'], **base_args),
    Extension('eelbrain._stats.adjacency_opt', [f'eelbrain/_stats/adjacency_opt{ext}'], **base_args),
    Extension('eelbrain._stats.opt', [f'eelbrain/_stats/opt{ext}'], **base_args),
    Extension('eelbrain._stats.vector', [f'eelbrain/_stats/vector{ext_cpp}'], include_dirs=['dsyevh3C'], **base_args),
]
if cythonize:
    extensions = cythonize(extensions)

setup(
    include_dirs=[np.get_include()],
    packages=find_packages(),
    ext_modules=extensions,
    scripts=['bin/eelbrain'],
)



import os
import timeit

import mne
from eelbrain import datasets, save

mne.set_log_level('warning')

fname = 'temp.pickle'
if not os.path.exists(fname):
    ds = datasets.get_mne_sample(-0.1, 0.2, src='ico', sub="modality == 'A'")
    source = ds['src'].source
    y = ds['src'][0].x
    save.pickle((y, source), fname)


setup = '''
from itertools import izip
import numpy as np
import scipy as sp
from eelbrain.lab import stats, load

y, source = load.unpickle(%r)

out = np.empty(y.shape, np.uint32)
bin_buff = np.empty(y.shape, np.bool_)
int_buff = np.empty(y.shape, np.uint32)
threshold = 1
tail = 0
struct = sp.ndimage.generate_binary_structure(y.ndim, 1)
struct[::2] = False
all_adjacent = False
flat_shape = (y.shape[0], np.prod(y.shape[1:]))
adjacency_src, adjacency_dst = source.adjacency().T
conn = {src:[] for src in np.unique(adjacency_src)}
for src, dst in izip(adjacency_src, adjacency_dst):
    conn[src].append(dst)
criteria=None
''' % fname

stmt = '''
stats.testnd._label_clusters(y, out, bin_buff, int_buff, threshold, tail, struct, all_adjacent, flat_shape, conn, criteria)
'''

timer = timeit.Timer(stmt, setup)
times = timer.repeat(100, 1)
print(times)
print("min = %s" % min(times))
print("avg of 10 lowest = %s" % (sum(sorted(times)[:10]) / 10))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import product
import os

from eelbrain.testing import TempDir
from eelbrain._experiment import TreeModel, FileTree


class Tree(TreeModel):
    _templates = dict(apath="/{afield}/",
                      afield=('a1', 'a2', 'a3'),
                      field2=('', 'value'))

    def __init__(self, **kwargs):
        TreeModel.__init__(self, **kwargs)
        self._register_compound('cmp', ('afield', 'field2'))


def test_tree():
    "Test simple formatting in the tree"
    tree = Tree()
    assert tree.get('apath') == '/a1/'
    vs = []
    for v in tree.iter('afield'):
        vs.append(v)
        assert tree.get('apath') == '/%s/' % v
        tree.set(afield='a3')
        assert tree.get('afield') == 'a3'
        assert tree.get('apath') == '/a3/'

    assert vs == ['a1', 'a2', 'a3']
    assert tree.get('afield') == 'a1'

    # test compound
    assert tree.get('cmp') == 'a1'
    tree.set(field2='value')
    assert tree.get('cmp') == 'a1_value'
    tree.set(field2='')
    assert tree.get('cmp') == 'a1'

    # test temporary state
    with tree._temporary_state:
        tree.set(afield='a2')
        assert tree.get('afield') == 'a2'
    assert tree.get('afield') == 'a1'

    # iterate
    assert list(tree.iter('afield')) == ['a1', 'a2', 'a3']
    assert list(tree.iter(('afield', 'field2'))) == [('a1', ''), ('a1', 'value'), ('a2', ''), ('a2', 'value'), ('a3', ''), ('a3', 'value')]
    assert list(tree.iter('cmp')) == ['a1', 'a1_value', 'a2', 'a2_value', 'a3', 'a3_value']
    assert list(tree.iter(('afield', 'cmp'))) == [('a1', 'a1'), ('a1', 'a1_value'), ('a2', 'a2'), ('a2', 'a2_value'), ('a3', 'a3'), ('a3', 'a3_value')]


class SlaveTree(TreeModel):
    _templates = {'path': '{a}_{b}_{sb_comp}_{slave}'}

    def __init__(self, a_seq, b_seq, c_seq):
        TreeModel.__init__(self)
        self._register_field('a', a_seq)
        self._register_field('b', b_seq, allow_empty=True)
        self._register_field('c', c_seq)
        self._register_compound('ab', ('a', 'b'))
        self._register_slave_field('s', 'a', lambda f: f['a'].upper())
        self._register_compound('sb', ('s', 'b'))
        self._register_slave_field('comp_slave', 'sb', lambda f: f['sb'].upper())
        # compound involving slave field
        self._register_field('s_a', a_seq, depends_on='c', slave_handler=self._update_sa)
        self._register_field('s_b', b_seq, depends_on='c', slave_handler=self._update_sb, allow_empty=True)
        self._register_compound('s_ab', ('s_a', 's_b'))
        self._store_state()

    @staticmethod
    def _update_sa(fields):
        if fields['c'] == 'c1':
            return 'a1'
        else:
            return 'a2'

    @staticmethod
    def _update_sb(fields):
        if fields['c'] == 'c1':
            return 'b1'
        else:
            return 'b2'


def test_slave_tree():
    a_seq = ['a1', 'a2', 'a3']
    b_seq = ['b1', 'b2', '']
    c_seq = ['c1', 'c2']
    ab_seq = [f'{a}_{b}' if b else a for a, b in product(a_seq, b_seq)]
    tree = SlaveTree(a_seq, b_seq, c_seq)

    # set
    assert tree.get('a') == 'a1'
    tree.set(a='a2')
    assert tree.get('a') == 'a2'
    tree.set(ab='a1 b2')
    assert tree.get('a') == 'a1'
    assert tree.get('b') == 'b2'
    tree.set(ab='a3')
    assert tree.get('a') == 'a3'
    assert tree.get('b') == ''

    tree.reset()
    assert tree.get('ab') == 'a1_b1'
    assert tree.get('sb') == 'A1_b1'
    assert tree.get('comp_slave') == 'A1_B1'
    tree.set(a='a2')
    assert tree.get('ab') == 'a2_b1'
    assert tree.get('sb') == 'A2_b1'
    assert tree.get('comp_slave') == 'A2_B1'

    # compound involving slave field
    tree.set(c='c2')
    assert tree.get('s_ab') == 'a2_b2'
    tree.set(c='c1')
    assert tree.get('s_ab') == 'a1_b1'

    # finde terminal keys
    assert tree.find_keys('c') == ['c']
    assert tree.find_keys('ab') == ['a', 'b']

    # .iter()
    assert list(tree.iter('a')) == a_seq
    assert list(tree.iter(('a', 'b'))) == list(product(a_seq, b_seq))
    assert list(tree.iter(('b', 'a'))) == list(product(b_seq, a_seq))
    # iter compound
    assert list(tree.iter('ab')) == ab_seq
    assert list(tree.iter(('c', 'ab'))) == list(product(c_seq, ab_seq))
    assert list(tree.iter('ab', values={'b': ''})) == a_seq
    assert list(tree.iter('ab', b='')) == a_seq


def test_file_tree():
    "Test file management tree"
    class Tree(FileTree):
        _templates = {'a-folder': '{root}/{folder}',
                      'a-file': '{a-folder}/{name}.txt',
                      'folder': ('f1', 'f2'),
                      'name': ('a1', 'a2', 'a3')}

        def __init__(self, *args, **kwargs):
            FileTree.__init__(self, *args, **kwargs)
            self._bind_make('a-file', self._make_a)

        def load_a(self):
            with open(self.get('a-file', make=True)) as fid:
                return fid.read()

        def _make_a(self):
            with open(self.get('a-file', mkdir=True), 'w') as fid:
                fid.write(self.format("{folder} {name}"))

    root = TempDir()
    tree = Tree(root=root)
    for _ in tree.iter_temp('a-file'):
        assert tree.load_a() == tree.format("{folder} {name}")

    for i, fname in enumerate(tree.iter_temp('a-file')):
        assert os.path.exists(fname)
    assert i == 5

    assert tree._glob_pattern('a-file', True, folder='f1') == f'{root}/f1/*.txt'

    tree.rm('a-file', name='*', folder='f1', confirm=True)
    for fname in tree.iter_temp('a-file', folder='f1'):
        assert fname[-6:-4] == tree.get('name')
        assert not os.path.exists(fname)
    for fname in tree.iter_temp('a-file', folder='f2'):
        assert fname[-6:-4] == tree.get('name')
        assert os.path.exists(fname)



"""Statistical tests for multidimensional data in :class:`NDVar` objects"""
__test__ = False
from ._stats.testnd import NDTest, MultiEffectNDTest, Correlation, TTestOneSample, TTestIndependent, TTestRelated, TContrastRelated, ANOVA, Vector, VectorDifferenceRelated
from ._stats.spm import LM, LMGroup
